from gwproactor.actors.actor import Actor, SyncThreadActor, SyncThreadT
from gwproactor.actors.web_event_listener import WebEventListener

__all__ = [
    "Actor",
    "SyncThreadActor",
    "SyncThreadT",
    "WebEventListener",
]



import asyncio
import uuid
from http import HTTPStatus
from typing import Any, Optional

from aiohttp.web_request import Request
from aiohttp.web_response import Response
from gwproto import Message
from gwproto.messages import AnyEvent, EventBase, ProblemEvent
from gwproto.named_types import SpaceheatNodeGt
from gwproto.named_types.web_server_gt import DEFAULT_WEB_SERVER_NAME
from pydantic import BaseModel, ValidationError
from result import Result

from gwproactor.actors.actor import Actor
from gwproactor.proactor_interface import AppInterface
from gwproactor.problems import Problems

EVENT_PATH: str = "/events"


class WebEventListenerSettings(BaseModel):
    http_path: str = EVENT_PATH
    server_name: str = DEFAULT_WEB_SERVER_NAME


class WebEventListener(Actor):
    DEFAULT_NODE_NAME: str = "web-event"

    def __init__(
        self,
        name: str,
        services: AppInterface,
        settings: Optional[WebEventListenerSettings] = None,
    ) -> None:
        super().__init__(name, services)
        if settings is None:
            settings = WebEventListenerSettings()
        self.services.add_web_route(
            server_name=settings.server_name,
            method="POST",
            path=settings.http_path,
            handler=self._handle_web_post,
        )

    async def _handle_web_post(self, request: Request) -> Response:
        status = HTTPStatus.OK
        response_text = ""
        problem_event: Optional[ProblemEvent] = None
        try:
            text = await request.text()
            try:
                event = AnyEvent.model_validate_json(text)
            except ValidationError as validation_error:
                status = HTTPStatus.UNPROCESSABLE_ENTITY
                response_text = validation_error.json(indent=2)
                problem_event = Problems(errors=[validation_error]).problem_event(
                    summary=f"ERROR decoding event <{self.name}>: <{response_text}>"
                )
            else:
                await asyncio.to_thread(self._wait_for_event_processing, event)
        except Exception as e:  # noqa: BLE001
            status = HTTPStatus.INTERNAL_SERVER_ERROR
            problem_event = Problems(errors=[e]).problem_event(
                summary=(
                    f"ERROR handling event post <{self.name}>: " f"{type(e)} <{e}>"
                )
            )
        if problem_event is not None:
            self.services.send_threadsafe(Message(Payload=problem_event))
        return Response(status=status, body=response_text)

    def _wait_for_event_processing(self, event: EventBase) -> None:
        self.services.wait_for_processing_threadsafe(
            Message(Payload=event),
        )

    def process_message(self, message: Message[Any]) -> Result[bool, Exception]:
        raise NotImplementedError("WebEventListener does not process internal messages")

    def start(self) -> None:
        """IOLoop will take care of start."""

    def stop(self) -> None:
        """IOLoop will take care of stop."""

    async def join(self) -> None:
        """IOLoop will take care of shutting down the associated task."""

    @classmethod
    def default_node(
        cls, name: str = DEFAULT_NODE_NAME, parent_name: str = "s"
    ) -> SpaceheatNodeGt:
        return SpaceheatNodeGt(
            ShNodeId=str(uuid.uuid4()),
            Name=name,
            ActorHierarchyName=f"{parent_name}.{name}",
            ActorClass=cls.__name__,
        )



import asyncio
import inspect
import logging
import textwrap
import time
from inspect import getframeinfo, stack
from pathlib import Path
from types import TracebackType
from typing import Any, Awaitable, Callable, Optional, Self, Type, Union

import typing_extensions

Predicate = Callable[[], bool]
AwaitablePredicate = Callable[[], Awaitable[bool]]
ErrorStringFunction = Callable[[], str]


class StopWatch:
    """Measure time with context manager"""

    start: float = 0
    end: float = 0
    elapsed: float = 0

    def __enter__(self) -> Self:
        self.start = time.time()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> typing_extensions.Literal[False]:
        self.end = time.time()
        self.elapsed = self.end - self.start
        return False


async def await_for(  # noqa: C901, PLR0912, PLR0913
    f: Union[Predicate, AwaitablePredicate],
    timeout: float,  # noqa: ASYNC109
    tag: str = "",
    raise_timeout: bool = True,  # noqa: FBT001, FBT002
    retry_duration: float = 0.01,
    err_str_f: Optional[ErrorStringFunction] = None,
    logger: Optional[logging.Logger | logging.LoggerAdapter[logging.Logger]] = None,
    error_dict: Optional[dict[str, Any]] = None,
) -> bool:
    """Similar to wait_for(), but awaitable. Instead of sleeping after a False resoinse from function f, await_for
    will asyncio.sleep(), allowing the event loop to continue. Additionally, f may be either a function or a coroutine.
    """
    now = start = time.time()
    until = now + timeout
    err_format = (
        "ERROR from {file}:{line}  [{tag}]\n"
        "  await_for() timed out after {seconds} seconds\n"
        "  wait function: {f}"
        "{err_str}"
    )
    if err_str_f is not None:

        def err_str_f_() -> str:
            return "\n" + textwrap.indent(err_str_f(), "  ")

    else:

        def err_str_f_() -> str:
            return ""

    f_is_async = inspect.iscoroutinefunction(f)
    result = False
    if now >= until:
        if f_is_async:
            result = await f()  # type:ignore[misc]
        else:
            result = f()  # type:ignore[assignment]
    while now < until and not result:
        if f_is_async:
            result = await f()  # type:ignore[misc]
        else:
            result = f()  # type:ignore[assignment]
        if not result:
            now = time.time()
            if now < until:
                await asyncio.sleep(min(retry_duration, until - now))
                now = time.time()
                # oops! we overslept
                if now >= until:
                    if f_is_async:
                        result = await f()  # type:ignore[misc]
                    else:
                        result = f()  # type:ignore[assignment]
    if result:
        return True
    caller = getframeinfo(stack()[1][0])
    format_dict = {
        "tag": tag,
        "file": Path(caller.filename).name,
        "line": caller.lineno,
        "seconds": time.time() - start,
        "f": f,
        "p": f(),
        "err_str": err_str_f_(),
    }
    err_str = err_format.format(**format_dict)
    if error_dict is not None:
        error_dict.update(
            format_dict,
            err_str=err_str,
        )
    if logger is not None:
        logger.error(err_str)
    if raise_timeout:
        raise ValueError(err_str)
    return False


def wait_for(
    f: Callable[[], bool],
    timeout: float,
    tag: str = "",
    raise_timeout: bool = True,  # noqa: FBT001, FBT002
    retry_duration: float = 0.1,
) -> bool:
    """Call function f() until it returns True or a timeout is reached. For async tests use await await_for() instead.
    retry_duration specified the sleep time between calls. If the timeout is reached before f return True, the function
    will either raise a ValueError (the default), or, if raise_timeout==False, it will return False. Function f is
    guaranteed to be called at least once. If an exception is raised the tag string will be attached to its message.
    """
    now = time.time()
    until = now + timeout
    if now >= until and f():
        return True
    while now < until:
        if f():
            return True
        now = time.time()
        if now < until:
            time.sleep(min(retry_duration, until - now))
            now = time.time()
    if raise_timeout:
        raise ValueError(
            f"ERROR. Function {f} timed out after {timeout} seconds. {tag}"
        )
    return False



from pathlib import Path
from typing import Optional


class PersisterException(Exception):
    path: Optional[Path] = None
    uid: str = ""

    def __init__(
        self, msg: str = "", uid: str = "", path: Optional[Path] = None
    ) -> None:
        self.path = path
        self.uid = uid
        super().__init__(msg)

    def __str__(self) -> str:
        s = self.__class__.__name__
        super_str = super().__str__()
        if super_str:
            s += f" [{super_str}]"
        s += f"  for uid: {self.uid}  path:{self.path}"
        return s


class PersisterError(PersisterException): ...


class PersisterWarning(PersisterException): ...


class WriteFailed(PersisterError): ...


class ContentTooLarge(PersisterError): ...


class FileMissing(PersisterError): ...


class ReadFailed(PersisterError): ...


class TrimFailed(PersisterError): ...


class ReindexError(PersisterError): ...


class DecodingError(PersisterError): ...


class ByteDecodingError(DecodingError): ...


class JSONDecodingError(DecodingError): ...


class EventDecodingError(DecodingError): ...


class UIDExistedWarning(PersisterWarning): ...


class FileExistedWarning(PersisterWarning): ...


class FileMissingWarning(PersisterWarning): ...


class UIDMissingWarning(PersisterWarning): ...


class FileEmptyWarning(PersisterWarning): ...



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, Optional, Sequence

from gwproto import Message

from gwproactor.message import MQTTReceiptPayload


@dataclass
class ReuploadCounts:
    started: int = 0
    completed: int = 0

    def start(self) -> None:
        self.started += 1

    def complete(self) -> None:
        self.completed += 1


@dataclass
class LinkStats:
    name: str
    num_received_by_type: dict[str, int] = field(
        default_factory=lambda: defaultdict(int)
    )
    num_received_by_topic: dict[str, int] = field(
        default_factory=lambda: defaultdict(int)
    )
    comm_event_counts: dict[str, int] = field(default_factory=lambda: defaultdict(int))
    reupload_counts: ReuploadCounts = field(default_factory=ReuploadCounts)
    timeouts: int = 0

    def start_reupload(self) -> None:
        self.reupload_counts.start()

    def complete_reupload(self) -> None:
        self.reupload_counts.complete()

    @property
    def num_received(self) -> int:
        return self.num_received_by_type[Message.type_name()]

    def __str__(self) -> str:
        s = f"LinkStats [{self.name}]  num_received: {self.num_received}  timeouts: {self.timeouts}"
        if self.num_received_by_type:
            s += "\n  Received by message_type:"
            for message_type in sorted(self.num_received_by_type):
                s += f"\n    {self.num_received_by_type[message_type]:3d}: [{message_type}]"
        if self.num_received_by_topic:
            s += "\n  Received by topic:"
            for topic in sorted(self.num_received_by_topic):
                s += f"\n    {self.num_received_by_topic[topic]:3d}: [{topic}]"
        if self.comm_event_counts:
            s += "\n  Comm event counts:"
            for comm_event in self.comm_event_counts:
                s += f"\n    {self.comm_event_counts[comm_event]:3d}: [{comm_event}]"
            s += f"\n    {self.reupload_counts.started:3d}: [reuploads_started]"
            s += f"\n    {self.reupload_counts.completed:3d}: [reuploads_completed]"
        return s


class ProactorStats:
    num_received_by_type: dict[str, int]
    num_received_by_topic: dict[str, int]
    num_events_received: int = 0
    links: dict[str, LinkStats]

    def __init__(self, link_names: Optional[Sequence[str]] = None) -> None:
        self.num_received_by_type = defaultdict(int)
        self.num_received_by_topic = defaultdict(int)
        if link_names is None:
            link_names = []
        self.links = {}
        for link_name in link_names:
            self.add_link(link_name)

    def add_message(self, message: Message[Any]) -> None:
        self.num_received_by_type[message.Header.MessageType] += 1

    def add_mqtt_message(self, message: Message[MQTTReceiptPayload]) -> None:
        self.num_received_by_topic[message.Payload.message.topic] += 1  # noqa
        link_stats = self.link(message.Payload.client_name)
        link_stats.num_received_by_type[Message.type_name()] += 1
        link_stats.num_received_by_type[message.Header.MessageType] += 1
        link_stats.num_received_by_topic[message.Payload.message.topic] += 1  # noqa
        if "gridworks-event" in message.Payload.message.topic:
            self.num_events_received += 1

    def add_decoded_mqtt_message_type(
        self, link_name: str, decoded_message_type: str
    ) -> None:
        self.link(link_name).num_received_by_type[decoded_message_type] += 1

    def total_received(self, message_type: str) -> int:
        return self.num_received_by_type.get(message_type, 0)

    @property
    def num_received(self) -> int:
        return self.num_received_by_type[Message.type_name()]

    @classmethod
    def make_link(cls, link_name: str) -> LinkStats:
        return LinkStats(link_name)

    def add_link(self, link_name: str) -> None:
        if link_name in self.links:
            raise ValueError(
                f"ERROR. link name {link_name} already present in self.links: {self.links.keys()}"
            )
        self.links[link_name] = self.make_link(link_name)

    def has_link(self, name: str) -> bool:
        return name in self.links

    def link(self, name: str) -> LinkStats:
        return self.links[name]

    def __str__(self) -> str:
        s = "ProactorStats Stats\n"
        if self.num_received_by_type:
            s += "\nGlobal received by message_type:"
            for message_type in sorted(self.num_received_by_type):
                s += f"\n    {self.num_received_by_type[message_type]:3d}: [{message_type}]"
            s += f"\n    {self.num_events_received:3d}: [gridworks.event*]"
        for link_name in sorted(self.links):
            s += "\n"
            s += str(self.links[link_name])
        return s



import asyncio
import copy
from collections import defaultdict
from typing import Any

from aiohttp import web
from aiohttp.typedefs import Handler as HTTPHandler
from aiohttp.web_routedef import RouteDef
from gwproto import Message
from gwproto.type_helpers import WebServerGt
from result import Result

from gwproactor.proactor_interface import AppInterface, Communicator, Runnable


class _RunWebServer:
    config: WebServerGt
    routes: list[RouteDef]

    def __init__(
        self,
        config: WebServerGt,
        routes: list[RouteDef],
    ) -> None:
        self.config = config
        self.routes = routes.copy()

    async def __call__(self) -> None:
        app = web.Application()
        app.add_routes(self.routes)
        runner = web.AppRunner(app)
        try:
            await runner.setup()
            site = web.TCPSite(
                runner,
                host=self.config.Host,
                port=self.config.Port,
                **self.config.Kwargs,
            )
            await site.start()
            while True:  # noqa: ASYNC110
                await asyncio.sleep(10)
        finally:
            try:  # noqa: SIM105
                await runner.cleanup()
            except:  # noqa: E722, S110
                pass


class _WebManager(Communicator, Runnable):
    _configs: dict[str, WebServerGt]
    _routes: dict[str, list[RouteDef]]

    def __init__(self, services: AppInterface) -> None:
        super().__init__("_WebManager", services)
        self._configs = {}
        self._routes = defaultdict(list)

    def process_message(self, message: Message[Any]) -> Result[bool, Exception]:  # noqa: ARG002
        raise ValueError("_WebManager does not currently process any messages")

    def disable(self) -> None:
        self._configs.clear()
        self._routes.clear()

    def start(self) -> None:
        for server_name, server_config in self._configs.items():
            if server_config.Enabled:
                self._services.io_loop_manager.add_io_coroutine(
                    _RunWebServer(
                        config=server_config,
                        routes=self._routes.get(server_name, []),
                    )(),
                    name=f"{self.name}.{server_name}",
                )

    def stop(self) -> None:
        pass

    async def join(self) -> None:
        pass

    def add_web_server_config(
        self,
        *,
        name: str,
        host: str,
        port: int,
        enabled: bool,
        server_kwargs: dict[str, Any],
    ) -> None:
        if name in self._configs:
            raise ValueError(f"ERROR: Server with name '{name}' already exists")
        self._configs[name] = WebServerGt(
            Name=name, Host=host, Port=port, Enabled=enabled, Kwargs=server_kwargs
        )

    def add_web_route(
        self,
        server_name: str,
        method: str,
        path: str,
        handler: HTTPHandler,
        **kwargs: Any,
    ) -> None:
        self._routes[server_name].append(
            RouteDef(
                method=method,
                path=path,
                handler=handler,
                kwargs=kwargs,
            )
        )

    def get_route_strings(self) -> dict[str, list[str]]:
        return {
            config_name: [str(route) for route in self._routes[config_name]]
            for config_name in self._routes
        }

    def get_configs(self) -> dict[str, WebServerGt]:
        return copy.deepcopy(self._configs)



import asyncio
import traceback
from asyncio import Task
from typing import Any, Iterable, Optional


def str_tasks(
    loop_: asyncio.AbstractEventLoop,
    tag: str = "",
    tasks: Optional[Iterable[Task[Any]]] = None,
) -> str:
    s = ""
    try:
        if tasks is None:
            tasks = asyncio.all_tasks(loop_)
        tasks = list(tasks)
        s += f"Tasks: {len(tasks)}  [{tag}]\n"

        def _get_task_exception(task_: asyncio.Task[Any]) -> BaseException | None:
            try:
                exception_ = task_.exception()
            except asyncio.CancelledError as _e:
                exception_ = _e
            except asyncio.InvalidStateError:
                exception_ = None
            return exception_

        for i, task in enumerate(tasks):
            s += (
                f"\t{i + 1}/{len(tasks)}  "
                f"{task.get_name():20s}  "
                f"done:{task.done()}   "
                f"exception:{_get_task_exception(task)}  "
                f"{task.get_coro()}\n"
            )
    except Exception as e:  # noqa: BLE001
        try:
            s += "ERROR in str_tasks:\n"
            s += "".join(traceback.format_exception(e))
            s += "\n"
        except:  # noqa: E722, S110
            pass
    return s



import typing
from collections import defaultdict

import rich
from gwproto import HardwareLayout, Message

from gwproactor import App, AppInterface, AppSettings
from gwproactor.actors.actor import PrimeActor
from gwproactor.config import MQTTClient
from gwproactor.config.links import CodecSettings, LinkSettings
from gwproactor.config.proactor_config import ProactorName
from gwproactor.message import MQTTReceiptPayload
from gwproactor.persister import TimedRollingFilePersister
from gwproactor_test.dummies import DUMMY_SCADA1_NAME, DUMMY_SCADA2_NAME
from gwproactor_test.dummies.names import DUMMY_ADMIN_NAME, DUMMY_ADMIN_SHORT_NAME
from gwproactor_test.dummies.tree.admin_messages import (
    AdminCommandSetRelay,
    AdminSetRelayEvent,
)
from gwproactor_test.dummies.tree.codecs import ScadaCodecFactory
from gwproactor_test.dummies.tree.messages import (
    RelayInfo,
    RelayReportEvent,
    SetRelay,
)


class DummyScada2(PrimeActor):
    ADMIN_LINK: str = "admin_link"

    relays: dict[str, bool]

    def __init__(self, name: str, services: AppInterface) -> None:
        super().__init__(name, services)
        self.relays = self.relays = defaultdict(bool)

    @property
    def admin_client(self) -> str:
        return self.ADMIN_LINK

    @classmethod
    def get_codec_factory(cls) -> ScadaCodecFactory:
        return ScadaCodecFactory()

    def _process_set_relay(self, payload: RelayInfo) -> None:
        self.services.logger.path(
            f"++{self.name}._process_set_relay "
            f"{payload.RelayName}  "
            f"closed:{payload.Closed}"
        )
        path_dbg = 0
        last_val = self.relays[payload.RelayName]
        event = RelayReportEvent(
            relay_name=payload.RelayName,
            closed=payload.Closed,
            changed=last_val != payload.Closed,
        )
        if event.changed:
            path_dbg |= 0x00000001
            self.relays[payload.RelayName] = event.closed
        self.services.generate_event(event)
        self.services.logger.path(
            f"--{self.name}._process_set_relay  "
            f"path:0x{path_dbg:08X}  "
            f"{int(last_val)} -> "
            f"{int(event.closed)}"
        )

    def _process_upstream_mqtt_message(
        self, message: Message[MQTTReceiptPayload], decoded: Message[typing.Any]
    ) -> None:
        self.services.logger.path(
            f"++{self.name}._process_downstream_mqtt_message {message.Payload.message.topic}",
        )
        path_dbg = 0
        match decoded.Payload:
            case SetRelay():
                path_dbg |= 0x00000001
                self._process_set_relay(decoded.Payload)
            case _:
                path_dbg |= 0x00000002
                rich.print(decoded.Header)
                raise ValueError(
                    f"There is no handler for mqtt message payload type [{type(decoded.Payload)}]\n"
                    f"Received\n\t topic: [{message.Payload.message.topic}]"
                )
        self.services.logger.path(
            f"--{self.name}._process_downstream_mqtt_message  path:0x{path_dbg:08X}",
        )

    def _process_admin_mqtt_message(
        self, message: Message[MQTTReceiptPayload], decoded: Message[typing.Any]
    ) -> None:
        self.services.logger.path(
            f"++{self.name}._process_admin_mqtt_message {message.Payload.message.topic}",
        )
        path_dbg = 0
        match decoded.Payload:
            case AdminCommandSetRelay() as command:
                path_dbg |= 0x00000001
                self.services.generate_event(AdminSetRelayEvent(command=command))
                self._process_set_relay(command.RelayInfo)
            case _:
                raise ValueError(
                    "In this test, since the environment is controlled, "
                    "there is no handler for mqtt message payload type "
                    f"[{type(decoded.Payload)}]\n"
                    f"Received\n\t topic: [{message.Payload.message.topic}]"
                )

        self.services.logger.path(
            f"--{self.name}._process_admin_mqtt_message  path:0x{path_dbg:08X}",
        )

    def process_mqtt_message(
        self, message: Message[MQTTReceiptPayload], decoded: Message[typing.Any]
    ) -> None:
        self.services.logger.path(
            f"++{self.name}._derived_process_mqtt_message {message.Payload.message.topic}",
        )
        path_dbg = 0
        if message.Payload.client_name == self.services.upstream_client:
            path_dbg |= 0x00000001
            self._process_upstream_mqtt_message(message, decoded)
        elif message.Payload.client_name == self.admin_client:
            path_dbg |= 0x00000002
            self._process_admin_mqtt_message(message, decoded)
        else:
            rich.print(decoded.Header)
            raise ValueError(
                "In this test, since the environment is controlled, "
                "there is no mqtt handler for message from client "
                f"[{message.Payload.client_name}]\n"
                f"Received\n\t topic: [{message.Payload.message.topic}]"
            )
        self.services.logger.path(
            f"--{self.name}._derived_process_mqtt_message  path:0x{path_dbg:08X}",
        )


class DummyScada2Settings(AppSettings):
    dummy_scada1: MQTTClient = MQTTClient()
    dummy_admin: MQTTClient = MQTTClient()


class DummyScada2App(App):
    SCADA1_LINK: str = DUMMY_SCADA1_NAME
    ADMIN_LINK: str = DUMMY_ADMIN_NAME

    @classmethod
    def app_settings_type(cls) -> type[DummyScada2Settings]:
        return DummyScada2Settings

    @classmethod
    def prime_actor_type(cls) -> type[DummyScada2]:
        return DummyScada2

    @property
    def prime_actor(self) -> DummyScada2:
        return typing.cast(DummyScada2, self._prime_actor)

    @classmethod
    def paths_name(cls) -> str:
        return DUMMY_SCADA2_NAME

    def _get_name(self, layout: HardwareLayout) -> ProactorName:
        return ProactorName(
            long_name=layout.scada_g_node_alias + ".s2",
            short_name="s2",
        )

    def _get_link_settings(
        self,
        name: ProactorName,  # noqa: ARG002
        layout: HardwareLayout,
        brokers: dict[str, MQTTClient],  # noqa: ARG002
    ) -> dict[str, LinkSettings]:
        return {
            self.SCADA1_LINK: LinkSettings(
                broker_name=self.SCADA1_LINK,
                peer_long_name=layout.scada_g_node_alias,
                peer_short_name="s",
                upstream=True,
                codec=CodecSettings(
                    message_modules=["gwproactor_test.dummies.tree.messages"]
                ),
            ),
            self.ADMIN_LINK: LinkSettings(
                broker_name=self.ADMIN_LINK,
                peer_long_name=DUMMY_ADMIN_NAME,
                peer_short_name=DUMMY_ADMIN_SHORT_NAME,
                link_subscription_short_name=layout.scada_g_node_alias + ".s2",
            ),
        }

    @classmethod
    def _make_persister(cls, settings: AppSettings) -> TimedRollingFilePersister:
        return TimedRollingFilePersister(settings.paths.event_dir)



import typer
from trogon import Trogon
from typer.main import get_group

from gwproactor.config import Paths
from gwproactor_test import persister_cli
from gwproactor_test.certs import generate_dummy_certs
from gwproactor_test.dummies.tree import admin_cli, atn1_cli, scada1_cli, scada2_cli
from gwproactor_test.dummies.tree.admin_settings import (
    AdminLinkSettings,
    DummyAdminSettings,
)
from gwproactor_test.dummies.tree.atn import DummyAtnApp
from gwproactor_test.dummies.tree.scada1 import DummyScada1App
from gwproactor_test.dummies.tree.scada2 import DummyScada2App

app = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    rich_markup_mode="rich",
    help="GridWorks Proactor Test CLI",
)

app.add_typer(scada1_cli.app, name="scada1", help="Use dummy scada1")
app.add_typer(scada2_cli.app, name="scada2", help="Use dummy scada1")
app.add_typer(atn1_cli.app, name="atn", help="Use dummy scada1")
app.add_typer(admin_cli.app, name="admin", help="Use dummy admin")
app.add_typer(persister_cli.app, name="mp", help="Measure persister timings.")


@app.command()
def gen_dummy_certs(dry_run: bool = False, only: str = "") -> None:
    """Generate certs for dummy proactors."""
    for app_name, settings in [
        ("atn", DummyAtnApp(env_file=DummyAtnApp.default_env_path()).settings),
        ("scada1", DummyScada1App(env_file=DummyScada1App.default_env_path()).settings),
        ("scada2", DummyScada2App(env_file=DummyScada2App.default_env_path()).settings),
        (
            "admin",
            DummyAdminSettings(
                _env_file=Paths.default_env_path(  # noqa
                    name=AdminLinkSettings.DUMMY_ADMIN_PATHS_NAME
                )
            ),
        ),  # noqa
    ]:
        if only and only != app_name:
            continue
        generate_dummy_certs(settings=settings, dry_run=dry_run)


@app.command()
def commands(ctx: typer.Context) -> None:
    """CLI command builder."""
    Trogon(get_group(app), click_context=ctx).run()


@app.callback()
def main_app_callback() -> None: ...


# For sphinx:
typer_click_object = typer.main.get_command(app)

if __name__ == "__main__":
    app()



import asyncio
import contextlib
import logging
import shutil
import typing
from pathlib import Path
from types import TracebackType
from typing import Optional, Self, Type

from pydantic_settings import BaseSettings

from gwproactor import AppSettings, Proactor, setup_logging
from gwproactor.app import App
from gwproactor.config import DEFAULT_BASE_NAME as DEFAULT_LOG_BASE_NAME
from gwproactor.config import MQTTClient, Paths
from gwproactor_test import copy_keys
from gwproactor_test.certs import uses_tls
from gwproactor_test.clean import TEST_HARDWARE_LAYOUT_PATH
from gwproactor_test.dummies.pair.child import DummyChildApp
from gwproactor_test.dummies.pair.parent import DummyParentApp
from gwproactor_test.instrumented_proactor import InstrumentedProactor
from gwproactor_test.logger_guard import LoggerGuards


def get_option_value(
    *,
    parameter_value: Optional[bool],
    option_name: str,
    request: Optional[typing.Any],
) -> bool:
    if parameter_value is not None or request is None:
        return bool(parameter_value)
    return bool(request.config.getoption(option_name))


class LiveTest:
    _parent_app: App
    _child_app: App
    verbose: bool
    child_verbose: bool
    parent_verbose: bool
    parent_on_screen: bool
    lifecycle_logging: bool
    logger_guards: LoggerGuards

    def __init__(
        self,
        *,
        child_app_settings: Optional[AppSettings] = None,
        parent_app_settings: Optional[AppSettings] = None,
        verbose: Optional[bool] = None,
        child_verbose: Optional[bool] = None,
        parent_verbose: Optional[bool] = None,
        lifecycle_logging: bool = False,
        add_child: bool = False,
        add_parent: bool = False,
        start_child: bool = False,
        start_parent: bool = False,
        parent_on_screen: Optional[bool] = None,
        request: typing.Any = None,
    ) -> None:
        self.verbose = get_option_value(
            parameter_value=verbose,
            option_name="--live-test-verbose",
            request=request,
        )
        self.child_verbose = get_option_value(
            parameter_value=child_verbose,
            option_name="--child-verbose",
            request=request,
        )
        self.parent_verbose = get_option_value(
            parameter_value=parent_verbose,
            option_name="--parent-verbose",
            request=request,
        )
        self.parent_on_screen = get_option_value(
            parameter_value=parent_on_screen,
            option_name="--parent-on-screen",
            request=request,
        )
        self.lifecycle_logging = lifecycle_logging
        self._child_app = self._make_app(
            self.child_app_type(),
            child_app_settings,
            app_verbose=self.child_verbose,
        )
        self._parent_app = self._make_app(
            self.parent_app_type(),
            parent_app_settings,
            app_verbose=self.parent_verbose,
        )
        self.setup_logging()
        if add_child or start_child:
            self.add_child()
            if start_child:
                self.start_child()
        if add_parent or start_parent:
            self.add_parent()
            if start_parent:
                self.start_parent()

    @classmethod
    def child_app_type(cls) -> type[App]:
        return DummyChildApp

    @property
    def child_app(self) -> App:
        return self._child_app

    @classmethod
    def parent_app_type(cls) -> type[App]:
        return DummyParentApp

    @property
    def parent_app(self) -> App:
        return self._parent_app

    def _make_app(
        self,
        app_type: type[App],
        app_settings: Optional[AppSettings],
        *,
        app_verbose: bool = False,
    ) -> App:
        # Copy hardware layout file.
        paths = Paths(name=app_type.paths_name())
        paths.mkdirs(parents=True, exist_ok=True)
        shutil.copyfile(Path(TEST_HARDWARE_LAYOUT_PATH), paths.hardware_layout)
        # Use an instrumented proactor
        sub_types = app_type.make_subtypes()
        sub_types.proactor_type = InstrumentedProactor
        app_settings = app_type.update_settings_from_command_line(
            app_type.get_settings(settings=app_settings, paths=paths),
            verbose=self.verbose or app_verbose,
        )
        if not self.lifecycle_logging and not self.verbose and not app_verbose:
            app_settings.logging.levels.lifecycle = logging.WARNING
        if app_settings.logging.base_log_name == str(DEFAULT_LOG_BASE_NAME):
            app_settings.logging.base_log_name = f"{DEFAULT_LOG_BASE_NAME}-{paths.name}"

        # Create the app
        app = app_type(
            paths=paths,
            app_settings=app_settings
            if app_settings is None
            else app_settings.with_paths(paths=paths),
            sub_types=sub_types,
        )
        # Copy keys.
        if uses_tls(app.config.settings):
            copy_keys(
                str(app.config.settings.paths.name),
                app.config.settings,
            )
        return app

    @property
    def parent(self) -> InstrumentedProactor:
        if self.parent_app.proactor is None:
            raise RuntimeError(
                "ERROR. CommTestHelper.parent accessed before creating parent."
                "pass add_parent=True to CommTestHelper constructor or call "
                "CommTestHelper.add_parent()"
            )
        return typing.cast(InstrumentedProactor, self.parent_app.proactor)

    @property
    def child(self) -> InstrumentedProactor:
        if self.child_app.raw_proactor is None:
            raise RuntimeError(
                "ERROR. CommTestHelper.child accessed before creating child."
                "pass add_child=True to CommTestHelper constructor or call "
                "CommTestHelper.add_child()"
            )
        return typing.cast(InstrumentedProactor, self.child_app.proactor)

    def start_child(
        self,
    ) -> Self:
        if self.child_app.raw_proactor is None:
            self.add_child()
        return self.start_proactor(self.child)

    def start_parent(
        self,
    ) -> Self:
        if self.parent_app.raw_proactor is None:
            self.add_parent()
        return self.start_proactor(self.parent)

    def start_proactor(self, proactor: Proactor) -> Self:
        asyncio.create_task(proactor.run_forever(), name=f"{proactor.name}_run_forever")  # noqa: RUF006
        return self

    def start(
        self,
    ) -> Self:
        return self

    def add_child(
        self,
    ) -> Self:
        self.child_app.instantiate()
        return self

    def add_parent(
        self,
    ) -> Self:
        self.parent_app.instantiate()
        return self

    def remove_child(
        self,
    ) -> Self:
        self.child_app.raw_proactor = None
        return self

    def remove_parent(
        self,
    ) -> Self:
        self.parent_app.raw_proactor = None
        return self

    @classmethod
    def _get_clients_supporting_tls(cls, settings: BaseSettings) -> list[MQTTClient]:
        clients = []
        for field_name in settings.__pydantic_fields__:
            v = getattr(settings, field_name)
            if isinstance(v, MQTTClient):
                clients.append(v)
        return clients

    def _get_child_clients_supporting_tls(self) -> list[MQTTClient]:
        """Overide to filter which MQTT clients of ChildSettingsT are treated as supporting TLS"""
        return self._get_clients_supporting_tls(self.child_app.config.settings)

    def _get_parent_clients_supporting_tls(self) -> list[MQTTClient]:
        """Overide to filter which MQTT clients of ParentSettingsT are treated as supporting TLS"""
        return self._get_clients_supporting_tls(self.parent_app.config.settings)

    @classmethod
    def _set_settings_use_tls(cls, use_tls: bool, clients: list[MQTTClient]) -> None:
        for client in clients:
            client.tls.use_tls = use_tls

    def set_use_tls(self, use_tls: bool) -> None:
        """Set MQTTClients which support TLS in parent and child settings to use TLS per use_tls. Clients supporting TLS
        is determined by _get_child_clients_supporting_tls() and _get_parent_clients_supporting_tls() which may be
        overriden in derived class.
        """
        self._set_settings_use_tls(use_tls, self._get_child_clients_supporting_tls())
        self._set_settings_use_tls(use_tls, self._get_parent_clients_supporting_tls())

    def setup_logging(self) -> None:
        child_settings = self.child_app.config.settings
        parent_settings = self.parent_app.config.settings
        child_settings.paths.mkdirs(parents=True)
        parent_settings.paths.mkdirs(parents=True)
        errors: list[Exception] = []
        if not self.lifecycle_logging and not self.verbose:
            if not self.child_verbose:
                child_settings.logging.levels.lifecycle = logging.WARNING
            if not self.parent_verbose:
                parent_settings.logging.levels.lifecycle = logging.WARNING
        self.logger_guards = LoggerGuards(
            list(child_settings.logging.qualified_logger_names().values())
            + list(parent_settings.logging.qualified_logger_names().values())
        )
        setup_logging(
            child_settings,
            errors=errors,
            add_screen_handler=True,
            root_gets_handlers=False,
        )
        assert not errors
        setup_logging(
            parent_settings,
            errors=errors,
            add_screen_handler=self.parent_on_screen,
            root_gets_handlers=False,
        )
        assert not errors

    def get_proactors(self) -> list[InstrumentedProactor]:
        proactors = []
        if self.child_app.raw_proactor is not None:
            proactors.append(self.child)
        if self.parent_app.raw_proactor is not None:
            proactors.append(self.parent)
        return proactors

    async def stop_and_join(self) -> None:
        proactors = self.get_proactors()
        for proactor in proactors:
            with contextlib.suppress(Exception):
                proactor.stop()
        for proactor in proactors:
            with contextlib.suppress(Exception):
                await proactor.join()

    async def __aenter__(
        self,
    ) -> Self:
        return self

    def get_log_path_str(self, exc: BaseException) -> str:
        return (
            f"CommTestHelper caught error {exc}.\n"
            "Working log dirs:"
            f"\n\t{self.child_app.config.settings.paths.log_dir}"
            f"\n\t{self.parent_app.config.settings.paths.log_dir}"
        )

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> bool:  # noqa
        try:
            await self.stop_and_join()
        finally:
            if exc is not None:
                try:
                    s = self.get_log_path_str(exc)
                except Exception as e:  # noqa: BLE001
                    try:
                        s = (
                            f"Caught {type(e)} / <{e}> while logging "
                            f"{type(exc)} / <{exc}>"
                        )
                    except:  # noqa: E722
                        s = "ERRORs upon errors in CommTestHelper cleanup"
                with contextlib.suppress(Exception):
                    logging.getLogger("gridworks").error(s)
            with contextlib.suppress(Exception):
                self.logger_guards.restore()  # noqa
        return False

    def summary_str(self) -> str:
        s = ""
        if self.child_app.raw_proactor is not None:
            s += "CHILD:\n" f"{self.child.summary_str()}\n"
        else:
            s += "CHILD: None\n"
        if self.parent_app.raw_proactor is not None:
            s += "PARENT:\n" f"{self.parent.summary_str()}"
        else:
            s += "PARENT: None\n"
        return s

    def assert_child_events_at_rest(
        self, *args: typing.Any, **kwargs: typing.Any
    ) -> None:
        self.child.assert_event_counts(*args, **kwargs)

    def assert_child1_events_at_rest(
        self, *args: typing.Any, **kwargs: typing.Any
    ) -> None:
        self.child.assert_event_counts(*args, **kwargs)

    def assert_parent_events_at_rest(
        self, *args: typing.Any, **kwargs: typing.Any
    ) -> None:
        self.child.assert_event_counts(*args, **kwargs)



from gwproto import HardwareLayout, MQTTCodec, create_message_model

from gwproactor import LinkSettings
from gwproactor.codecs import CodecFactory
from gwproactor.config.proactor_config import ProactorName
from gwproactor_test.dummies.names import DUMMY_ADMIN_NAME
from gwproactor_test.dummies.tree.admin_messages import (
    AdminCommandReadRelays,
    AdminCommandSetRelay,
)


class AdminCodec(MQTTCodec):
    def __init__(self) -> None:
        super().__init__(
            create_message_model(
                model_name="AdminMessageDecoder",
                explicit_types=[AdminCommandSetRelay, AdminCommandReadRelays],
            )
        )

    def validate_source_and_destination(self, src: str, dst: str) -> None: ...


class ScadaCodecFactory(CodecFactory):
    def get_codec(
        self,
        link_name: str,
        link: LinkSettings,
        proactor_name: ProactorName,
        layout: HardwareLayout,  # noqa: ARG002
    ) -> MQTTCodec:
        if link_name == DUMMY_ADMIN_NAME:
            return AdminCodec()
        return super().get_codec(
            link_name=link_name,
            link=link,
            proactor_name=proactor_name,
            layout=layout,
        )



from typing import Optional

from result import Ok, Result

from gwproactor.persister.interface import PersisterInterface
from gwproactor.problems import Problems


class StubPersister(PersisterInterface):
    _num_persists: int = 0
    _num_retrieves: int = 0
    _num_clears: int = 0

    def persist(self, uid: str, content: bytes) -> Result[bool, Problems]:  # noqa: ARG002
        self._num_persists += 1
        return Ok()

    def clear(self, uid: str) -> Result[bool, Problems]:  # noqa: ARG002
        self._num_clears += 1
        return Ok()

    def pending_ids(self) -> list[str]:
        return []

    @property
    def num_pending(self) -> int:
        return 0

    @property
    def curr_bytes(self) -> int:
        return 0

    def __contains__(self, uid: str) -> bool:
        return False

    def retrieve(self, uid: str) -> Result[Optional[bytes], Problems]:  # noqa: ARG002
        return Ok(None)

    def reindex(self) -> Result[Optional[bool], Problems]:
        self._num_retrieves += 1
        return Ok()

    @property
    def num_persists(self) -> int:
        return self._num_persists

    @property
    def num_retrieves(self) -> int:
        return self._num_retrieves

    @property
    def num_clears(self) -> int:
        return self._num_clears



import os
from typing import Optional


class ExternalWatchdogCommandBuilder:
    """Create arguments which will be passed to subprocess.run() to pat the external watchdog.

    If the returned list is empty, pat process will be run.

    By default an empty list is returned if the environment variable named by sevice_variable_name() is not set to
    1 or true.
    """

    @classmethod
    def service_variable_name(cls, service_name: str) -> str:
        return f"{service_name.upper()}_RUNNING_AS_SERVICE"

    @classmethod
    def running_as_service(cls, service_name: str) -> bool:
        return os.getenv(cls.service_variable_name(service_name), "").lower() in [
            "1",
            "true",
        ]

    @classmethod
    def default_pat_args(cls, pid: Optional[int] = None) -> list[str]:  # noqa: ARG003
        return []

    @classmethod
    def pat_args(
        cls,
        service_name: str,
        args: Optional[list[str]] = None,
        pid: Optional[int] = None,
    ) -> list[str]:
        """Return arguments to be passed to subprocess.run() to pat the external watchdog."""
        if cls.running_as_service(service_name):
            if args is None:
                args = cls.default_pat_args(pid=pid)
            return args
        return []


class SystemDWatchdogCommandBuilder(ExternalWatchdogCommandBuilder):
    @classmethod
    def default_pat_args(cls, pid: Optional[int] = None) -> list[str]:
        if pid is None:
            pid = os.getpid()
        return [
            "systemd-notify",
            f"--pid={pid}",
            "WATCHDOG=1",
        ]



"""Scada implementation"""

import typing
from typing import Any

from gwproto import HardwareLayout, Message
from gwproto.messages import EventBase
from result import Ok, Result

from gwproactor import App, AppSettings
from gwproactor.actors.actor import PrimeActor
from gwproactor.config import MQTTClient
from gwproactor.config.links import LinkSettings
from gwproactor.config.proactor_config import ProactorName
from gwproactor.message import DBGPayload, MQTTReceiptPayload
from gwproactor.persister import (
    PersisterInterface,
    TimedRollingFilePersister,
)
from gwproactor_test.dummies import DUMMY_CHILD_NAME, DUMMY_PARENT_NAME


class DummyParent(PrimeActor):
    def process_internal_message(self, message: Message[Any]) -> None:
        self.process_message(message)

    def process_message(self, message: Message[Any]) -> Result[bool, Exception]:
        match message.Payload:
            case DBGPayload():
                message.Header.Src = self.services.publication_name
                dst_client = message.Header.Dst
                message.Header.Dst = ""
                self.services.publish_message(dst_client, message)
        return Ok(True)

    def process_mqtt_message(
        self, mqtt_client_message: Message[MQTTReceiptPayload], decoded: Message[Any]
    ) -> None:
        self.services.logger.path(
            f"++{self.name}.process_mqtt_message %s",
            mqtt_client_message.Payload.message.topic,
        )
        path_dbg = 0
        self.services.stats.add_message(decoded)
        match decoded.Payload:
            case EventBase():
                path_dbg |= 0x00000001
                self.services.generate_event(decoded.Payload)
            case _:
                path_dbg |= 0x00000002
        self.services.logger.path(
            f"--{self.name}.process_mqtt_message  path:0x%08X", path_dbg
        )


class DummyParentSettings(AppSettings):
    child: MQTTClient = MQTTClient()


class DummyParentApp(App):
    CHILD_MQTT: str = DUMMY_CHILD_NAME

    @classmethod
    def app_settings_type(cls) -> type[DummyParentSettings]:
        return DummyParentSettings

    @classmethod
    def prime_actor_type(cls) -> type[DummyParent]:
        return DummyParent

    @property
    def prime_actor(self) -> DummyParent:
        return typing.cast(DummyParent, self._prime_actor)

    @classmethod
    def paths_name(cls) -> str:
        return DUMMY_PARENT_NAME

    def _get_name(self, layout: HardwareLayout) -> ProactorName:
        return ProactorName(
            long_name=layout.atn_g_node_alias,
            short_name="a",
        )

    def _get_link_settings(
        self,
        name: ProactorName,  # noqa: ARG002
        layout: HardwareLayout,
        brokers: dict[str, MQTTClient],  # noqa: ARG002
    ) -> dict[str, LinkSettings]:
        return {
            self.CHILD_MQTT: LinkSettings(
                broker_name=self.CHILD_MQTT,
                peer_long_name=layout.scada_g_node_alias,
                peer_short_name="s",
                downstream=True,
            )
        }

    @classmethod
    def _make_persister(cls, settings: AppSettings) -> PersisterInterface:
        return TimedRollingFilePersister(settings.paths.event_dir)



import ssl
from pathlib import Path
from ssl import VerifyMode
from typing import Optional

from pydantic import BaseModel, SecretStr

from gwproactor.config.paths import TLSPaths


class TLSInfo(BaseModel):
    """TLS settings for a single MQTT client"""

    use_tls: bool = True
    port: int = 8883
    paths: TLSPaths = TLSPaths()
    cert_reqs: Optional[VerifyMode] = ssl.CERT_REQUIRED
    ciphers: Optional[str] = None
    keyfile_password: SecretStr = SecretStr("")

    def update_tls_paths(self, certs_dir: str | Path, client_name: str) -> "TLSInfo":
        """Calculate non-set paths given a certs_dir and client name. Meant to be called in context where those are
        known, e.g. a validator on a higher-level model which has access to a Paths object and a named MQTT
        configuration."""
        self.paths = self.paths.effective_paths(certs_dir, client_name)
        return self


class MQTTClient(BaseModel):
    """Settings for connecting to an MQTT Broker"""

    host: str = "localhost"
    port: int = 1883
    keepalive: int = 60
    bind_address: str = ""
    bind_port: int = 0
    username: Optional[str] = None
    password: SecretStr = SecretStr("")
    tls: TLSInfo = TLSInfo()

    def update_tls_paths(self, certs_dir: str | Path, client_name: str) -> "MQTTClient":
        """Calculate non-set paths given a certs_dir and client name. Meant to be called in context where those are
        known, e.g. a validator on a higher-level model which has access to a Paths object and a named MQTT
        configuration."""
        self.tls.update_tls_paths(certs_dir, client_name)
        return self

    def effective_port(self) -> int:
        if self.tls.use_tls:
            return self.tls.port
        return self.port



from gwproactor_test.dummies.names import (
    DUMMY_ATN_ENV_PREFIX,
    DUMMY_ATN_NAME,
    DUMMY_CHILD_ENV_PREFIX,
    DUMMY_CHILD_NAME,
    DUMMY_ENV_PREFIX,
    DUMMY_PARENT_ENV_PREFIX,
    DUMMY_PARENT_NAME,
    DUMMY_SCADA1_ENV_PREFIX,
    DUMMY_SCADA1_NAME,
    DUMMY_SCADA2_ENV_PREFIX,
    DUMMY_SCADA2_NAME,
)
from gwproactor_test.dummies.pair.child import DummyChildApp
from gwproactor_test.dummies.pair.parent import DummyParentApp

__all__ = [
    "DUMMY_CHILD_ENV_PREFIX",
    "DUMMY_CHILD_NAME",
    "DUMMY_ENV_PREFIX",
    "DUMMY_PARENT_ENV_PREFIX",
    "DUMMY_PARENT_NAME",
    "DummyChildApp",
    "DummyParentApp",
    "DUMMY_SCADA1_ENV_PREFIX",
    "DUMMY_SCADA2_ENV_PREFIX",
    "DUMMY_SCADA1_NAME",
    "DUMMY_SCADA2_NAME",
    "DUMMY_ATN_NAME",
    "DUMMY_ATN_ENV_PREFIX",
]



import textwrap
import traceback
from typing import Optional, Sequence

from gwproto.messages import ProblemEvent
from gwproto.messages import Problems as ProblemType

_EFFECTIVELY_NO_MAX = 1000000000


class Problems(ValueError):
    MAX_PROBLEMS = 10

    errors: list[BaseException]
    warnings: list[BaseException]
    max_problems: int = MAX_PROBLEMS

    def __init__(
        self,
        msg: str = "",
        warnings: Optional[Sequence[BaseException]] = None,
        errors: Optional[Sequence[BaseException]] = None,
        max_problems: Optional[int] = MAX_PROBLEMS,
    ) -> None:
        self.errors = [] if errors is None else list(errors)
        self.warnings = [] if warnings is None else list(warnings)
        if max_problems is None:
            max_problems = _EFFECTIVELY_NO_MAX
        self.max_problems = max_problems
        super().__init__(msg)

    def __bool__(self) -> bool:
        return bool(self.errors) or bool(self.warnings)

    def add_error(self, error: Exception) -> "Problems":
        if len(self.errors) < self.max_problems:
            self.errors.append(error)
        return self

    def add_warning(self, warning: Exception) -> "Problems":
        if len(self.warnings) < self.max_problems:
            self.warnings.append(warning)
        return self

    def add_problems(self, other: "Problems") -> "Problems":
        self.errors.extend(other.errors[: self.max_problems - len(self.errors)])
        self.warnings.extend(other.warnings[: self.max_problems - len(self.warnings)])
        return self

    def __str__(self) -> str:
        if bool(self):
            s = f"Problems: {len(self.errors)} errors, {len(self.warnings)} warnings, max: {self.max_problems}"
            for attr_name in ["errors", "warnings"]:
                lst = getattr(self, attr_name)
                if lst:
                    s += f"\n{attr_name.capitalize()}:\n"
                    for i, entry in enumerate(lst):
                        s += f"  {i:2d}: <{entry.__class__.__name__}>  " f"<{entry}>\n"
            return s
        return ""

    def error_traceback_str(self) -> str:
        s = ""
        for i, error in enumerate(self.errors):
            s += f"Traceback for error {i + 1} / {len(self.errors)}:\n"
            for line in traceback.format_exception(error):
                s += textwrap.indent(line, "  ")
        return s

    def __repr__(self) -> str:
        return str(self)

    def problem_event(self, summary: str, src: str = "") -> ProblemEvent:
        return ProblemEvent(
            Src=src,
            ProblemType=ProblemType.error if self.errors else ProblemType.warning,
            Summary=summary,
            Details=str(self),
        )

    def __len__(self) -> int:
        return len(self.errors) + len(self.warnings)



from gwproactor.persister.exceptions import (
    ByteDecodingError,
    ContentTooLarge,
    DecodingError,
    FileEmptyWarning,
    FileExistedWarning,
    FileMissing,
    FileMissingWarning,
    JSONDecodingError,
    PersisterError,
    PersisterException,
    PersisterWarning,
    ReadFailed,
    ReindexError,
    TrimFailed,
    UIDExistedWarning,
    UIDMissingWarning,
    WriteFailed,
)
from gwproactor.persister.interface import PersisterInterface
from gwproactor.persister.simple_directory_writer import SimpleDirectoryWriter
from gwproactor.persister.stub import StubPersister
from gwproactor.persister.timed_rolling_file import TimedRollingFilePersister

__all__ = [
    "ByteDecodingError",
    "ContentTooLarge",
    "DecodingError",
    "FileEmptyWarning",
    "FileExistedWarning",
    "FileMissing",
    "FileMissingWarning",
    "JSONDecodingError",
    "PersisterError",
    "PersisterException",
    "PersisterInterface",
    "PersisterWarning",
    "ReadFailed",
    "ReindexError",
    "SimpleDirectoryWriter",
    "StubPersister",
    "TimedRollingFilePersister",
    "TrimFailed",
    "UIDExistedWarning",
    "UIDMissingWarning",
    "WriteFailed",
]



import uuid
from typing import Literal

from gwproto.messages import EventBase
from pydantic import BaseModel, Field

from gwproactor_test.dummies.tree.messages import RelayInfo


class AdminInfo(BaseModel):
    User: str
    SrcMachine: str


class AdminCommandSetRelay(BaseModel):
    CommandInfo: AdminInfo
    RelayInfo: RelayInfo
    MessageId: str = Field(default_factory=lambda: str(uuid.uuid4()))
    TypeName: Literal["gridworks.dummy.admin.command.set.relay"] = (
        "gridworks.dummy.admin.command.set.relay"
    )


class AdminCommandReadRelays(BaseModel):
    CommandInfo: AdminInfo
    MessageId: str = Field(default_factory=lambda: str(uuid.uuid4()))
    TypeName: Literal["gridworks.dummy.admin.command.read.relays"] = (
        "gridworks.dummy.admin.command.read.relays"
    )


class AdminSetRelayEvent(EventBase):
    command: AdminCommandSetRelay
    TypeName: Literal["gridworks.event.admin.command.set.relay"] = (
        "gridworks.event.admin.command.set.relay"
    )



from pydantic_settings import BaseSettings, SettingsConfigDict

MQTT_LINK_POLL_SECONDS = 60.0
ACK_TIMEOUT_SECONDS = 5.0
NUM_INITIAL_EVENT_REUPLOADS: int = 5
NUM_INFLIGHT_EVENTS: int = 50


class ProactorSettings(BaseSettings):
    mqtt_link_poll_seconds: float = MQTT_LINK_POLL_SECONDS
    ack_timeout_seconds: float = ACK_TIMEOUT_SECONDS
    num_initial_event_reuploads: int = NUM_INITIAL_EVENT_REUPLOADS
    num_inflight_events: int = NUM_INFLIGHT_EVENTS

    model_config = SettingsConfigDict(
        env_prefix="PROACTOR_",
        env_nested_delimiter="__",
        nested_model_default_partial_update=True,
    )



from pydantic_settings import BaseSettings

# 	publish 	MY_LONG_NAME/to/PEER_SHORT_NAME
# 	subscribe	PEER_LONG_NAME/to/MY_SHORT_NAME
#
# 	subscription topic:
# 		src: peer gnode_name/long_name
# 		dst: link subscription_name or link mgr subscription_name
#
#
# 	publication topic
# 		src: self.long_name or message.Header.Src
# 		dst: peer.short_name or message.Header.Dst


class CodecSettings(BaseSettings):
    message_model_name: str = ""
    message_modules: list[str] = []
    use_default_message_modules: bool = True


class LinkSettings(BaseSettings):
    broker_name: str = ""
    enabled: bool = True
    peer_long_name: str = ""
    peer_short_name: str = ""
    link_subscription_short_name: str = ""
    codec: CodecSettings = CodecSettings()
    upstream: bool = False
    downstream: bool = False



"""Code for actors that use a simple rest interaction, converting the response to one or more
REST commands into a message posted to main processing thread.

"""

import asyncio
from dataclasses import dataclass, field
from typing import Any, Awaitable, Callable, Optional

import aiohttp
import yarl
from aiohttp import ClientResponse, ClientSession, ClientTimeout
from gwproto import Message
from gwproto.data_classes.components.rest_poller_component import RESTPollerComponent
from gwproto.type_helpers import AioHttpClientTimeout, RESTPollerSettings, URLConfig
from result import Ok, Result

from gwproactor import Actor, AppInterface, Problems
from gwproactor.proactor_interface import INVALID_IO_TASK_HANDLE, IOLoopInterface

Converter = Callable[[ClientResponse], Awaitable[Optional[Message[Any]]]]
ThreadSafeForwarder = Callable[[Message[Any]], Any]


async def null_converter(_: ClientResponse) -> Optional[Message[Any]]:
    return None


def null_forwarder(_: Message[Any]) -> None:
    return None


def to_client_timeout(
    timeout: Optional[AioHttpClientTimeout],
) -> Optional[ClientTimeout]:
    if timeout is not None:
        return ClientTimeout(**timeout.model_dump())
    return None


@dataclass
class SessionArgs:
    base_url: Optional[yarl.URL] = None
    kwargs: dict[str, Any] = field(default_factory=dict)


@dataclass
class RequestArgs:
    method: str = "GET"
    url: Optional[yarl.URL] = None
    kwargs: dict[str, Any] = field(default_factory=dict)


class RESTPoller:
    _name: str
    _rest: RESTPollerSettings
    _io_loop_manager: IOLoopInterface
    _task_id: int
    _session_args: Optional[SessionArgs] = None
    _request_args: Optional[RequestArgs] = None
    _converter: Converter
    _forward: ThreadSafeForwarder

    def __init__(  # noqa: PLR0913
        self,
        name: str,
        rest: RESTPollerSettings,
        loop_manager: IOLoopInterface,
        *,
        convert: Converter = null_converter,
        forward: ThreadSafeForwarder = null_forwarder,
        cache_request_args: bool = True,
    ) -> None:
        self._name = name
        self._task_id = INVALID_IO_TASK_HANDLE
        self._rest = rest
        self._io_loop_manager = loop_manager
        setattr(self, "_convert", convert)  # noqa: B010
        self._forward = forward
        if cache_request_args:
            self._session_args = self._make_session_args()
            self._request_args = self._make_request_args()

    def _make_base_url(self) -> Optional[yarl.URL]:
        return URLConfig.make_url(self._rest.session.base_url)

    def _make_url(self) -> Optional[yarl.URL]:
        return URLConfig.make_url(self._rest.request.url)

    def _make_session_args(self) -> SessionArgs:
        return SessionArgs(
            self._make_base_url(),
            dict(
                self._rest.session.model_dump(
                    exclude={"base_url", "timeout"},
                    exclude_unset=True,
                ),
                timeout=to_client_timeout(self._rest.session.timeout),
            ),
        )

    def _make_request_args(self) -> RequestArgs:
        return RequestArgs(
            self._rest.request.method,
            self._make_url(),
            dict(
                self._rest.request.model_dump(
                    exclude={"method", "url", "timeout"},
                    exclude_unset=True,
                ),
                timeout=to_client_timeout(self._rest.request.timeout),
            ),
        )

    async def _make_request(self, session: ClientSession) -> Optional[ClientResponse]:
        try:
            args = self._request_args
            if args is None:
                args = self._make_request_args()
            if args.url is None:
                raise ValueError(f"ERROR. args <{args}> produced url of None")  # noqa: TRY301
            response = await session.request(args.method, args.url, **args.kwargs)
            if self._rest.errors.request.error_for_http_status:
                response.raise_for_status()
        except Exception as e:
            response = None
            if self._rest.errors.request.report:
                try:  # noqa: SIM105
                    self._forward(
                        Message(
                            Payload=Problems(errors=[e]).problem_event(
                                summary=(
                                    f"Request error for <{self._name}>: {type(e)} <{e}>"
                                ),
                            )
                        )
                    )
                except:  # noqa: E722, S110
                    pass
            if self._rest.errors.request.raise_exception:
                raise
        return response

    async def _convert(self, response: ClientResponse) -> Optional[Message[Any]]:
        try:
            message = await self._converter(response)
        except Exception as convert_exception:
            message = None
            if self._rest.errors.convert.report:
                try:  # noqa: SIM105
                    self._forward(
                        Message(
                            Payload=Problems(errors=[convert_exception]).problem_event(
                                summary=(
                                    f"Convert error for <{self._name}>: {type(convert_exception)} <{convert_exception}>"
                                ),
                            )
                        )
                    )
                except:  # noqa: E722, S110
                    pass
            if self._rest.errors.convert.raise_exception:
                raise
        return message

    def _get_next_sleep_seconds(self) -> float:
        return self._rest.poll_period_seconds

    async def _run(self) -> None:
        reconnect = True
        while reconnect:
            args = self._session_args
            if args is None:
                args = self._make_session_args()
            async with aiohttp.ClientSession(args.base_url, **args.kwargs) as session:
                while True:
                    response = await self._make_request(session)
                    if response is not None:
                        async with response:
                            message = await self._convert(response)
                        if message is not None:
                            self._forward(message)
                    sleep_seconds = self._get_next_sleep_seconds()
                    await asyncio.sleep(sleep_seconds)

    def start(self) -> None:
        self._task_id = self._io_loop_manager.add_io_coroutine(
            self._run(),
            name=self._name,
        )

    def stop(self) -> None:
        self._io_loop_manager.cancel_io_routine(self._task_id)


class RESTPollerActor(Actor):
    _poller: RESTPoller

    def __init__(
        self,
        name: str,
        services: AppInterface,
        *,
        convert: Converter = null_converter,
        forward: ThreadSafeForwarder = null_forwarder,
        cache_request_args: bool = True,
    ) -> None:
        super().__init__(name, services)
        component = services.hardware_layout.component(self.name)
        if not isinstance(component, RESTPollerComponent):
            display_name = getattr(
                component, "display_name", "MISSING ATTRIBUTE display_name"
            )
            raise TypeError(
                f"ERROR. Component <{display_name}> has type {type(component)}. "
                f"Expected RESTPollerComponent.\n"
                f"  Node: {self.name}\n"
                f"  Component id: {getattr(component, 'component_id', None)}"
            )
        self._poller = RESTPoller(
            name,
            component.rest,
            services.io_loop_manager,
            convert=convert,
            forward=forward,
            cache_request_args=cache_request_args,
        )

    def process_message(self, _: Message[Any]) -> Result[bool, Exception]:
        return Ok()

    def start(self) -> None:
        self._poller.start()

    def stop(self) -> None:
        self._poller.stop()

    async def join(self) -> None:
        """IOLoop will take care of shutting down the associated task."""



import typing
from types import ModuleType

import rich
from gwproto import HardwareLayout, Message
from gwproto.messages import EventBase

from gwproactor import App, AppInterface, AppSettings, actors
from gwproactor.actors.actor import PrimeActor
from gwproactor.config import MQTTClient
from gwproactor.config.links import CodecSettings, LinkSettings
from gwproactor.config.proactor_config import ProactorName
from gwproactor.message import MQTTReceiptPayload
from gwproactor.persister import TimedRollingFilePersister
from gwproactor_test.dummies import DUMMY_ATN_NAME, DUMMY_SCADA1_NAME, DUMMY_SCADA2_NAME
from gwproactor_test.dummies.names import DUMMY_ADMIN_NAME, DUMMY_ADMIN_SHORT_NAME
from gwproactor_test.dummies.tree.admin_messages import (
    AdminCommandReadRelays,
    AdminCommandSetRelay,
    AdminSetRelayEvent,
)
from gwproactor_test.dummies.tree.codecs import ScadaCodecFactory
from gwproactor_test.dummies.tree.messages import (
    RelayInfoReported,
    RelayReportEvent,
    RelayReportReceivedEvent,
    RelayStates,
    SetRelay,
    SetRelayMessage,
)


class DummyScada1(PrimeActor):
    relays: RelayStates

    def __init__(self, name: str, services: AppInterface) -> None:
        super().__init__(name, services)
        self.relays = RelayStates()

    def set_relay(self, relay_name: str, closed: bool) -> None:
        self.services.send_threadsafe(
            SetRelayMessage(
                src=self.name,
                dst=self.name,
                relay_name=relay_name,
                closed=closed,
            )
        )

    def process_internal_message(self, message: Message[typing.Any]) -> None:
        self.services.logger.path(
            f"++{self.name}._derived_process_message "
            f"{message.Header.Src}/{message.Header.MessageType}"
        )
        path_dbg = 0
        match message.Payload:
            case SetRelay():
                path_dbg |= 0x00000001
                self.services.publish_message(
                    self.services.downstream_client,
                    SetRelayMessage(
                        src=self.services.publication_name,
                        relay_name=message.Payload.RelayName,
                        closed=message.Payload.Closed,
                        ack_required=True,
                    ),
                )
            case _:
                path_dbg |= 0x00000002
        self.services.logger.path(
            "--{self.name}._derived_process_message  path:0x{path_dbg:08X}"
        )

    def _process_report_relay_event(self, event: RelayReportEvent) -> None:
        self.services.logger.path(
            f"++{self.name}._process_set_relay_event "
            f"{event.relay_name}  "
            f"closed:{event.closed}  "
            f"changed: {event.changed}"
        )
        path_dbg = 0
        if event.relay_name not in self.relays.Relays:
            self.relays.Relays[event.relay_name] = RelayInfoReported()
        last_val = self.relays.Relays[event.relay_name].Closed
        self.relays.Relays[event.relay_name].Closed = event.closed
        changed = last_val != self.relays.Relays[event.relay_name].Closed
        self.services.logger.info(
            f"{event.relay_name}:  {int(last_val)} -> "
            f"{int(self.relays.Relays[event.relay_name].Closed)}  "
            f"changed: {int(changed)}/{int(event.changed)}"
        )
        report_received_event = RelayReportReceivedEvent(
            relay_name=event.relay_name,
            closed=event.closed,
            changed=event.changed,
        )

        if changed != event.changed:
            path_dbg |= 0x00000001
            report_received_event.mismatch = True
            self.relays.Relays[event.relay_name].CurrentChangeMismatch = True
            self.relays.Relays[event.relay_name].MismatchCount += 1
            self.relays.TotalChangeMismatches += 1
            report_received_event.mismatch_count = self.relays.TotalChangeMismatches
            self.services.logger.info(
                f"State change mismatch for {event.relay_name}  "
                f"found: {int(changed)}  reported: {event.changed}  "
                f"total mismatches: {self.relays.TotalChangeMismatches}"
            )
        self.services.generate_event(report_received_event)
        self.services.logger.path(
            f"--{self.name}._process_set_relay_event "
            f"{event.relay_name}  "
            f"closed:{event.closed}  "
            f"changed: {event.changed}  "
            f"path: 0x{path_dbg:08X}"
        )

    def _process_event(self, event: EventBase) -> None:
        self.services.logger.path(
            f"++_process_event  {event.TypeName}  from:{event.Src}",
        )
        self.services.generate_event(event)
        if isinstance(event, RelayReportEvent):
            self._process_report_relay_event(event)
        self.services.logger.path("--_process_event")

    def _process_downstream_mqtt_message(
        self, message: Message[MQTTReceiptPayload], decoded: Message[typing.Any]
    ) -> None:
        self.services.logger.path(
            f"++{self.name}._process_downstream_mqtt_message {message.Payload.message.topic}",
        )
        path_dbg = 0
        match decoded.Payload:
            case EventBase():
                path_dbg |= 0x00000001
                self._process_event(decoded.Payload)
            case _:
                # For testing purposes, this should fail.
                path_dbg |= 0x00000002
                rich.print(decoded.Header)
                raise ValueError(
                    "In this test, since the environment is controlled, "
                    "there is no handler for mqtt message payload type "
                    f"[{type(decoded.Payload)}]\n"
                    f"Received\n\t topic: [{message.Payload.message.topic}]"
                )
        self.services.logger.path(
            f"--{self.name}._process_downstream_mqtt_message  path:0x{path_dbg:08X}",
        )

    def _process_admin_mqtt_message(
        self, message: Message[MQTTReceiptPayload], decoded: Message[typing.Any]
    ) -> None:
        self.services.logger.path(
            f"++{self.name}._process_admin_mqtt_message {message.Payload.message.topic}",
        )
        path_dbg = 0
        match decoded.Payload:
            case AdminCommandSetRelay() as command:
                path_dbg |= 0x00000001
                self.services.generate_event(AdminSetRelayEvent(command=command))
                self.services.publish_message(
                    self.services.downstream_client,
                    SetRelayMessage(
                        src=self.services.publication_name,
                        relay_name=command.RelayInfo.RelayName,
                        closed=command.RelayInfo.Closed,
                        ack_required=True,
                    ),
                )
            case AdminCommandReadRelays():
                path_dbg |= 0x00000002
                self.services.publish_message(
                    self.admin_client,
                    Message(
                        Src=self.services.publication_name,
                        Payload=self.relays.model_copy(),
                    ),
                )
            case _:
                raise ValueError(
                    "In this test, since the environment is controlled, "
                    "there is no handler for mqtt message payload type "
                    f"[{type(decoded.Payload)}]\n"
                    f"Received\n\t topic: [{message.Payload.message.topic}]"
                )

        self.services.logger.path(
            f"--{self.name}._process_admin_mqtt_message  path:0x{path_dbg:08X}",
        )

    def process_mqtt_message(
        self, message: Message[MQTTReceiptPayload], decoded: Message[typing.Any]
    ) -> None:
        self.services.logger.path(
            f"++{self.name}._derived_process_mqtt_message {message.Payload.message.topic}",
        )
        path_dbg = 0
        if message.Payload.client_name == self.services.downstream_client:
            path_dbg |= 0x00000001
            self._process_downstream_mqtt_message(message, decoded)
        elif message.Payload.client_name == self.admin_client:
            path_dbg |= 0x00000002
            self._process_admin_mqtt_message(message, decoded)
        else:
            rich.print(decoded.Header)
            raise ValueError(
                "In this test, since the environment is controlled, "
                "there is no mqtt handler for message from client "
                f"[{message.Payload.client_name}]\n"
                f"Received\n\t topic: [{message.Payload.message.topic}]"
            )
        self.services.logger.path(
            f"--{self.name}._derived_process_mqtt_message  path:0x{path_dbg:08X}",
        )

    @property
    def admin_client(self) -> str:
        return DUMMY_ADMIN_NAME

    @classmethod
    def get_codec_factory(cls) -> ScadaCodecFactory:
        return ScadaCodecFactory()


class DummyScada1Settings(AppSettings):
    dummy_atn1: MQTTClient = MQTTClient()
    dummy_scada2: MQTTClient = MQTTClient()
    dummy_admin: MQTTClient = MQTTClient()


class DummyScada1App(App):
    ATN_LINK: str = DUMMY_ATN_NAME
    SCADA2_LINK: str = DUMMY_SCADA2_NAME
    ADMIN_LINK: str = DUMMY_ADMIN_NAME

    @classmethod
    def app_settings_type(cls) -> type[DummyScada1Settings]:
        return DummyScada1Settings

    @classmethod
    def prime_actor_type(cls) -> type[DummyScada1]:
        return DummyScada1

    @property
    def prime_actor(self) -> DummyScada1:
        return typing.cast(DummyScada1, self._prime_actor)

    @classmethod
    def actors_module(cls) -> ModuleType:
        return actors

    @classmethod
    def paths_name(cls) -> str:
        return DUMMY_SCADA1_NAME

    def _get_name(self, layout: HardwareLayout) -> ProactorName:
        return ProactorName(
            long_name=layout.scada_g_node_alias,
            short_name="s",
        )

    def _get_link_settings(
        self,
        name: ProactorName,  # noqa: ARG002
        layout: HardwareLayout,
        brokers: dict[str, MQTTClient],  # noqa: ARG002
    ) -> dict[str, LinkSettings]:
        return {
            self.ATN_LINK: LinkSettings(
                broker_name=self.ATN_LINK,
                peer_long_name=layout.atn_g_node_alias,
                peer_short_name="a",
                upstream=True,
            ),
            self.SCADA2_LINK: LinkSettings(
                broker_name=self.SCADA2_LINK,
                peer_long_name=layout.scada_g_node_alias + ".s2",
                peer_short_name="s2",
                downstream=True,
                codec=CodecSettings(
                    message_modules=["gwproactor_test.dummies.tree.messages"]
                ),
            ),
            self.ADMIN_LINK: LinkSettings(
                broker_name=self.ADMIN_LINK,
                peer_long_name=DUMMY_ADMIN_NAME,
                peer_short_name=DUMMY_ADMIN_SHORT_NAME,
                link_subscription_short_name=layout.scada_g_node_alias,
            ),
        }

    @classmethod
    def _make_persister(cls, settings: AppSettings) -> TimedRollingFilePersister:
        return TimedRollingFilePersister(settings.paths.event_dir)



"""Message structures for use between proactor and its sub-objects."""

import uuid
from enum import Enum
from typing import Any, Generic, Literal, Optional, Sequence, TypeVar

from gwproto import as_enum
from gwproto.message import Header, Message, ensure_arg
from gwproto.messages import EventBase
from paho.mqtt.client import ConnectFlags, MQTTMessage
from paho.mqtt.reasoncodes import ReasonCode as PahoReasonCode
from pydantic import BaseModel, ConfigDict, field_validator

from gwproactor.config import LoggerLevels
from gwproactor.problems import Problems


class MessageType(Enum):
    invalid = "invalid"
    mqtt_subscribe = "mqtt_subscribe"
    mqtt_message = "mqtt_message"
    mqtt_connected = "mqtt_connected"
    mqtt_disconnected = "mqtt_disconnected"
    mqtt_connect_failed = "mqtt_connect_failed"
    mqtt_suback = "mqtt_suback"
    mqtt_problems = "mqtt_problems"


class KnownNames(Enum):
    proactor = "proactor"
    mqtt_clients = "mqtt_clients"
    watchdog_manager = "watchdog_manager"
    io_loop_manager = "io_loop_manager"


class MQTTClientsPayload(BaseModel):
    client_name: str
    userdata: Optional[Any] = None


MQTTClientsPayloadT = TypeVar("MQTTClientsPayloadT", bound=MQTTClientsPayload)


class MQTTClientMessage(Message[MQTTClientsPayloadT], Generic[MQTTClientsPayloadT]):
    def __init__(
        self,
        message_type: MessageType,
        payload: MQTTClientsPayloadT,
    ) -> None:
        super().__init__(
            Header=Header(
                Src=KnownNames.mqtt_clients.value,
                Dst=KnownNames.proactor.value,
                MessageType=message_type.value,
            ),
            Payload=payload,
        )


class MQTTMessageModel(BaseModel):
    timestamp: float = 0
    state: int = 0
    dup: bool = False
    mid: int = 0
    topic: str = ""
    payload: bytes = b""
    qos: int = 0
    retain: bool = False

    @classmethod
    def from_mqtt_message(cls, message: MQTTMessage) -> "MQTTMessageModel":
        model = MQTTMessageModel()
        for field_name in model.__pydantic_fields__:
            setattr(model, field_name, getattr(message, field_name))
        return model


class MQTTReceiptPayload(MQTTClientsPayload):
    message: MQTTMessageModel


class MQTTReceiptMessage(MQTTClientMessage[MQTTReceiptPayload]):
    def __init__(
        self,
        client_name: str,
        userdata: Optional[Any],
        message: MQTTMessage,
    ) -> None:
        super().__init__(
            message_type=MessageType.mqtt_message,
            payload=MQTTReceiptPayload(
                client_name=client_name,
                userdata=userdata,
                message=MQTTMessageModel.from_mqtt_message(message),
            ),
        )


class SerializedReasonCode(BaseModel):
    packet_type: int
    code: int
    string: str

    @classmethod
    def from_paho_reason_code(
        cls, reason_code: PahoReasonCode
    ) -> "SerializedReasonCode":
        return SerializedReasonCode(
            packet_type=reason_code.packetType,
            code=reason_code.value,
            string=str(reason_code),
        )


class MQTTSubackPayload(MQTTClientsPayload):
    mid: int
    reason_codes: Sequence[SerializedReasonCode]


class MQTTSubackMessage(MQTTClientMessage[MQTTSubackPayload]):
    def __init__(
        self,
        client_name: str,
        userdata: Optional[Any],
        mid: int,
        reason_codes: Sequence[PahoReasonCode],
    ) -> None:
        super().__init__(
            message_type=MessageType.mqtt_suback,
            payload=MQTTSubackPayload(
                client_name=client_name,
                userdata=userdata,
                mid=mid,
                reason_codes=[
                    SerializedReasonCode.from_paho_reason_code(reason_code)
                    for reason_code in reason_codes
                ],
            ),
        )


class MQTTCommEventPayload(MQTTClientsPayload):
    rc: Optional[SerializedReasonCode]


class MQTTConnectPayload(MQTTCommEventPayload):
    flags: ConnectFlags


class MQTTConnectMessage(MQTTClientMessage[MQTTConnectPayload]):
    def __init__(
        self,
        client_name: str,
        userdata: Optional[Any],
        flags: ConnectFlags,
        rc: PahoReasonCode,
    ) -> None:
        super().__init__(
            message_type=MessageType.mqtt_connected,
            payload=MQTTConnectPayload(
                client_name=client_name,
                userdata=userdata,
                flags=flags,
                rc=SerializedReasonCode.from_paho_reason_code(rc),
            ),
        )


class MQTTConnectFailPayload(MQTTClientsPayload):
    pass


class MQTTConnectFailMessage(MQTTClientMessage[MQTTConnectFailPayload]):
    def __init__(self, client_name: str, userdata: Optional[Any]) -> None:
        super().__init__(
            message_type=MessageType.mqtt_connect_failed,
            payload=MQTTConnectFailPayload(
                client_name=client_name,
                userdata=userdata,
            ),
        )


class MQTTDisconnectPayload(MQTTCommEventPayload):
    pass


class MQTTDisconnectMessage(MQTTClientMessage[MQTTDisconnectPayload]):
    def __init__(
        self, client_name: str, userdata: Optional[Any], rc: PahoReasonCode
    ) -> None:
        super().__init__(
            message_type=MessageType.mqtt_disconnected,
            payload=MQTTDisconnectPayload(
                client_name=client_name,
                userdata=userdata,
                rc=SerializedReasonCode.from_paho_reason_code(rc),
            ),
        )


class MQTTProblemsPayload(MQTTCommEventPayload):
    problems: Problems
    model_config = ConfigDict(arbitrary_types_allowed=True)


class MQTTProblemsMessage(MQTTClientMessage[MQTTCommEventPayload]):
    def __init__(
        self, client_name: str, problems: Problems, rc: Optional[PahoReasonCode] = None
    ) -> None:
        super().__init__(
            message_type=MessageType.mqtt_problems,
            payload=MQTTProblemsPayload(
                client_name=client_name,
                rc=SerializedReasonCode.from_paho_reason_code(rc)
                if rc is not None
                else rc,
                problems=problems,
            ),
        )


class PatWatchdog(BaseModel): ...


class PatInternalWatchdog(PatWatchdog):
    TypeName: Literal["gridworks.watchdog.pat.internal"] = (
        "gridworks.watchdog.pat.internal"
    )


class PatExternalWatchdog(PatWatchdog):
    TypeName: Literal["gridworks.watchdog.pat.external"] = (
        "gridworks.watchdog.pat.external"
    )


class PatInternalWatchdogMessage(Message[PatInternalWatchdog]):
    def __init__(self, src: str) -> None:
        super().__init__(
            Src=src,
            Dst=KnownNames.watchdog_manager.value,
            Payload=PatInternalWatchdog(),
        )


class PatExternalWatchdogMessage(Message[PatExternalWatchdog]):
    def __init__(self) -> None:
        super().__init__(
            Src=KnownNames.watchdog_manager.value,
            Dst=KnownNames.watchdog_manager.value,
            Payload=PatExternalWatchdog(),
        )


class Command(BaseModel): ...


CommandT = TypeVar("CommandT", bound=Command)


class CommandMessage(Message[CommandT], Generic[CommandT]):
    def __init__(self, *, AckRequired: bool = True, **kwargs: Any) -> None:
        ensure_arg("MessageId", str(uuid.uuid4()), kwargs)
        super().__init__(AckRequired=AckRequired, **kwargs)


class Shutdown(Command):
    Reason: str = ""
    TypeName: Literal["gridworks.shutdown"] = "gridworks.shutdown"


class ShutdownMessage(CommandMessage[Shutdown]):
    def __init__(self, *, Reason: str = "", **data: Any) -> None:
        ensure_arg("Payload", Shutdown(Reason=Reason), data)
        super().__init__(**data)


class InternalShutdownMessage(ShutdownMessage):
    def __init__(self, *, AckRequired: bool = False, **data: Any) -> None:
        super().__init__(AckRequired=AckRequired, **data)


class DBGCommands(Enum):
    show_subscriptions = "show_subscriptions"


class DBGPayload(BaseModel):
    Levels: LoggerLevels = LoggerLevels(
        message_summary=-1,
        lifecycle=-1,
        comm_event=-1,
    )
    Command: Optional[DBGCommands] = None
    TypeName: Literal["gridworks.proactor.dbg"] = "gridworks.proactor.dbg"

    @field_validator("Command", mode="before")
    @classmethod
    def command_value(cls, v: Any) -> Optional[DBGCommands]:
        return as_enum(v, DBGCommands)


class DBGEvent(EventBase):
    Command: DBGPayload
    Path: str = ""
    Count: int = 0
    Msg: str = ""
    TypeName: Literal["gridworks.event.proactor.dbg"] = "gridworks.event.proactor.dbg"



"""Proactor implementation"""

import asyncio
import sys
import threading
import traceback
from functools import cached_property
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
)

import gwproto
from aiohttp.typedefs import Handler as HTTPHandler
from gwproto import Message
from gwproto.data_classes.components.web_server_component import WebServerComponent
from gwproto.data_classes.hardware_layout import HardwareLayout
from gwproto.data_classes.sh_node import ShNode
from gwproto.messages import Ack, EventBase, EventT, Ping, ProblemEvent, ShutdownEvent
from gwproto.named_types.web_server_gt import WebServerGt
from paho.mqtt.client import MQTTMessageInfo
from result import Err, Ok, Result

from gwproactor.callbacks import (
    CallbackManager,
    ProactorCallbackFunctions,
    ProactorCallbackInterface,
)
from gwproactor.config.app_settings import AppSettings
from gwproactor.config.proactor_config import ProactorConfig, ProactorName
from gwproactor.external_watchdog import (
    ExternalWatchdogCommandBuilder,
    SystemDWatchdogCommandBuilder,
)
from gwproactor.io_loop import IOLoop
from gwproactor.links import (
    AckWaitInfo,
    AsyncioTimerManager,
    LinkManager,
)
from gwproactor.links.mqtt import QOS
from gwproactor.logger import ProactorLogger
from gwproactor.message import (
    DBGCommands,
    DBGEvent,
    DBGPayload,
    MQTTConnectFailPayload,
    MQTTConnectPayload,
    MQTTDisconnectPayload,
    MQTTProblemsPayload,
    MQTTReceiptPayload,
    MQTTSubackPayload,
    PatWatchdog,
    Shutdown,
)
from gwproactor.persister import PersisterInterface
from gwproactor.proactor_interface import (
    AppInterface,
    CommunicatorInterface,
    IOLoopInterface,
    MonitoredName,
    Runnable,
)
from gwproactor.problems import Problems
from gwproactor.stats import ProactorStats
from gwproactor.str_tasks import str_tasks
from gwproactor.watchdog import WatchdogManager
from gwproactor.web_manager import _WebManager

T = TypeVar("T")


class Proactor(AppInterface, Runnable):
    AWAIT_PROCESSING_FUTURE_ATTRIBUTE: str = "_await_processing_future"

    _name: ProactorName
    _settings: AppSettings
    _node: ShNode
    _callbacks: CallbackManager
    _layout: HardwareLayout
    _logger: ProactorLogger
    _stats: ProactorStats
    _event_persister: PersisterInterface
    _reindex_problems: Optional[Problems] = None
    _loop: Optional[asyncio.AbstractEventLoop] = None
    _receive_queue: Optional[asyncio.Queue[Any]] = None
    _processing_futures: set[asyncio.Future[Any]]
    _processing_futures_lock: asyncio.Lock
    _links: LinkManager
    _communicators: Dict[str, CommunicatorInterface]
    _stop_requested: bool
    _stopped: bool
    _tasks: List[asyncio.Task[Any]]
    _io_loop_manager: IOLoop
    _web_manager: _WebManager
    _watchdog: WatchdogManager

    def __init__(self, config: ProactorConfig) -> None:
        self._name = config.name
        self._settings = config.settings
        self._callbacks = CallbackManager(callback_functions=config.callback_functions)
        self._layout = config.layout
        self._node = self._layout.node(self.name)
        self._logger = config.logger
        self._stats = self.make_stats()
        self._event_persister = config.event_persister
        self._logger.lifecycle(f"Proactor <{self._name}> reindexing events")
        reindex_result = self._event_persister.reindex()
        self._logger.lifecycle(
            f"Proactor <{self._name}> reindexing complete.\n"
            f"  {self._event_persister.num_pending} events present for upload, "
            f"using approximately {int(self._event_persister.curr_bytes / 1024)} KB / "
            f"{round(self._event_persister.curr_bytes / 1024 / 1024, 1)} MB "
            f"storage space."
        )
        if reindex_result.is_err():
            self._reindex_problems = reindex_result.err()
            self._logger.error("ERROR in event persister reindex():")
            self._logger.error(reindex_result.err())
        self._links = LinkManager(
            publication_name=self.publication_name,
            subscription_name=self.subscription_name,
            settings=self._settings.proactor,
            logger=self._logger,
            stats=self._stats,
            event_persister=self._event_persister,
            timer_manager=AsyncioTimerManager(),
            ack_timeout_callback=self._process_ack_timeout,
        )
        self._processing_futures = set()
        self._processing_futures_lock = asyncio.Lock()
        self._communicators = {}
        self._tasks = []
        self._stop_requested = False
        self._stopped = False
        self._watchdog = WatchdogManager(9, self)
        self.add_communicator(self._watchdog)
        self._io_loop_manager = IOLoop(self)
        self.add_communicator(self._io_loop_manager)
        self._web_manager = _WebManager(self)
        self.add_communicator(self._web_manager)
        for server_config in self._layout.get_components_by_type(WebServerComponent):
            self._web_manager.add_web_server_config(
                name=server_config.web_server_gt.Name,
                host=server_config.web_server_gt.Host,
                port=server_config.web_server_gt.Port,
                enabled=server_config.web_server_gt.Enabled,
                server_kwargs=server_config.web_server_gt.Kwargs,
            )

    @classmethod
    def make_stats(cls) -> ProactorStats:
        return ProactorStats()

    def send(self, message: Message[Any]) -> None:
        if self._receive_queue is None:
            raise RuntimeError("ERROR. send() called before Proactor started.")
        if not isinstance(message.Payload, PatWatchdog):
            self._logger.message_summary(
                direction="OUT internal",
                src=message.Header.Src,
                dst=message.Header.Dst,
                topic=f"{message.Header.Src}/to/{message.Header.Dst}/{message.Header.MessageType}",
                payload_object=message.Payload,
                message_id=message.Header.MessageId,
            )
        self._receive_queue.put_nowait(message)

    async def _add_processing_future(self, future: asyncio.Future[Any]) -> None:
        async with self._processing_futures_lock:
            self._processing_futures.add(future)

    async def _clear_processing_future(self, future: asyncio.Future[Any]) -> None:
        async with self._processing_futures_lock:
            if future in self._processing_futures:
                self._processing_futures.remove(future)

    async def _clear_processing_futures(self) -> list[asyncio.Future[Any]]:
        futures = []
        async with self._processing_futures_lock:
            futures = list(self._processing_futures)
            self._processing_futures.clear()
        return futures

    async def _get_processing_future(self) -> asyncio.Future[Any]:
        if self._loop is None:
            raise RuntimeError(
                "ERROR. _get_processing_future() called before Proactor started."
            )
        future = self._loop.create_future()
        await self._add_processing_future(future)
        return future

    async def _attach_future_to_message(
        self, message: Message[Any]
    ) -> asyncio.Future[Any]:
        if self._loop is None:
            raise RuntimeError(
                "ERROR. await_processing() called before Proactor started."
            )
        future = getattr(message, self.AWAIT_PROCESSING_FUTURE_ATTRIBUTE, None)
        if future is None:
            future = await self._get_processing_future()
            setattr(message, self.AWAIT_PROCESSING_FUTURE_ATTRIBUTE, future)
        else:
            if not isinstance(future, asyncio.Future):
                raise RuntimeError(
                    "ERROR. await_processing() received message "
                    f"(type: {message.message_type()}) with "
                    f"{self.AWAIT_PROCESSING_FUTURE_ATTRIBUTE} attribute "
                    f"(type: {type(future)}) which is not an instance of "
                    f"asyncio.Future."
                )
            await self._add_processing_future(future)
        return future

    @classmethod
    async def _await_processing(
        cls, future: asyncio.Future[Result[Any, BaseException]]
    ) -> Result[Any, BaseException]:
        try:
            await future
        except asyncio.CancelledError as canceled:
            return Err(canceled)
        e = future.exception()
        if e is not None:
            return Err(e)
        return future.result()

    async def await_processing(
        self, message: Message[Any]
    ) -> Result[Any, BaseException]:
        if self._stop_requested:
            return Ok(value=False)
        future = await self._attach_future_to_message(message)
        self.send(message)
        return await self._await_processing(future)

    def send_threadsafe(self, message: Message[Any]) -> None:
        if self._loop is None or self._receive_queue is None:
            raise RuntimeError(
                "ERROR. send_threadsafe() called before Proactor started."
            )
        self._loop.call_soon_threadsafe(self._receive_queue.put_nowait, message)

    def wait_for_processing_threadsafe(
        self, message: Message[Any]
    ) -> Result[Any, BaseException]:
        if self._loop is None:
            raise RuntimeError(
                "ERROR. wait_for_send_threadsafe() called before Proactor started."
            )
        return asyncio.run_coroutine_threadsafe(
            self.await_processing(message), self._loop
        ).result()

    def get_communicator_names(self) -> set[str]:
        return set(self._communicators.keys())

    def get_communicator(self, name: str) -> Optional[CommunicatorInterface]:
        return self._communicators.get(name, None)

    def get_communicator_as_type(self, name: str, type_: Type[T]) -> Optional[T]:
        communicator = self.get_communicator(name)
        if communicator is not None and not isinstance(communicator, type_):
            raise ValueError(
                f"ERROR. Communicator <{name}> has type {type(communicator)} not {type_}"
            )
        return communicator

    @cached_property
    def name(self) -> str:
        return self._name.name

    @cached_property
    def long_name(self) -> str:
        return self._name.long_name

    @cached_property
    def short_name(self) -> str:
        return self._name.short_name

    @cached_property
    def paths_name(self) -> str:
        return str(self.settings.paths.name)

    @cached_property
    def publication_name(self) -> str:
        return self._name.publication_name

    @cached_property
    def subscription_name(self) -> str:
        return self._name.subscription_name

    @cached_property
    def name_object(self) -> ProactorName:
        return self._name

    @property
    def callback_functions(self) -> ProactorCallbackFunctions:
        return self._callbacks.callback_functions

    @property
    def monitored_names(self) -> Sequence[MonitoredName]:
        return []

    @property
    def settings(self) -> AppSettings:
        return self._settings

    @property
    def logger(self) -> ProactorLogger:
        return self._logger

    @property
    def stats(self) -> ProactorStats:
        return self._stats

    @property
    def links(self) -> LinkManager:
        return self._links

    def add_callbacks(self, callbacks: ProactorCallbackInterface) -> int:
        return self._callbacks.add_callbacks(callbacks)

    def remove_callbacks(self, callbacks_id: int) -> None:
        return self._callbacks.remove_callbacks(callbacks_id)

    def publish_message(  # noqa: PLR0913
        self,
        link_name: str,
        message: Message[Any],
        qos: int = 0,
        context: Any = None,
        *,
        topic: str = "",
        use_link_topic: bool = False,
    ) -> MQTTMessageInfo:
        return self._links.publish_message(
            link_name=link_name,
            message=message,
            qos=qos,
            context=context,
            topic=topic,
            use_link_topic=use_link_topic,
        )

    def publish_upstream(
        self, payload: Any, qos: QOS = QOS.AtMostOnce, **message_args: Any
    ) -> MQTTMessageInfo:
        return self._links.publish_upstream(payload=payload, qos=qos, **message_args)

    @property
    def event_persister(self) -> PersisterInterface:
        return self._event_persister

    @property
    def io_loop_manager(self) -> IOLoopInterface:
        return self._io_loop_manager

    @property
    def hardware_layout(self) -> HardwareLayout:
        return self._layout

    @property
    def upstream_client(self) -> str:
        return self._links.upstream_client

    @property
    def downstream_client(self) -> str:
        return self._links.downstream_client

    @property
    def async_receive_queue(self) -> Optional[asyncio.Queue[Any]]:
        return self._receive_queue

    @property
    def event_loop(self) -> Optional[asyncio.AbstractEventLoop]:
        return self._loop

    def add_web_server_config(
        self, name: str, host: str, port: int, **kwargs: Any
    ) -> None:
        self._web_manager.add_web_server_config(
            name=name, host=host, port=port, **kwargs
        )

    def add_web_route(
        self,
        server_name: str,
        method: str,
        path: str,
        handler: HTTPHandler,
        **kwargs: Any,
    ) -> None:
        self._web_manager.add_web_route(
            server_name=server_name, method=method, path=path, handler=handler, **kwargs
        )

    def get_web_server_route_strings(self) -> dict[str, list[str]]:
        return self._web_manager.get_route_strings()

    def get_web_server_configs(self) -> dict[str, WebServerGt]:
        return self._web_manager.get_configs()

    def get_external_watchdog_builder_class(
        self,
    ) -> type[ExternalWatchdogCommandBuilder]:
        return SystemDWatchdogCommandBuilder

    def _send(self, message: Message[Any]) -> None:
        self.send(message)

    def generate_event(self, event: EventT) -> Result[bool, Exception]:
        return self._links.generate_event(event)

    def _process_ack_timeout(self, wait_info: AckWaitInfo) -> None:
        self._logger.message_enter(
            "++Proactor<%s>._process_ack_timeout %s",
            self.short_name,
            wait_info.message_id,
        )
        path_dbg = 0
        match self._links.process_ack_timeout(wait_info):
            case Ok(transition):
                path_dbg |= 0x00000001
                if transition.deactivated():
                    path_dbg |= 0x00000002
                    self._callbacks.recv_deactivated(transition)
            case Err(exception):
                path_dbg |= 0x00000004
                self._report_error(exception, msg="Proactor._process_ack_timeout")
        self._logger.message_exit(
            "--Proactor<%s>._process_ack_timeout path:0x%08X", self.short_name, path_dbg
        )

    def _process_ack(self, link_name: str, message_id: str) -> None:
        self._links.process_ack(link_name, message_id)

    def _process_dbg(self, dbg: DBGPayload) -> None:
        self._logger.path("++_process_dbg")
        path_dbg = 0
        count_dbg = 0
        for logger_name in ["message_summary", "lifecycle", "comm_event"]:
            requested_level = getattr(dbg.Levels, logger_name)
            if requested_level > -1:
                path_dbg |= 0x00000001
                count_dbg += 1
                logger = getattr(self._logger, logger_name + "_logger")
                old_level = logger.getEffectiveLevel()
                logger.setLevel(requested_level)
                self._logger.debug(
                    "%s logger level %s -> %s",
                    logger_name,
                    old_level,
                    logger.getEffectiveLevel(),
                )
        match dbg.Command:
            case DBGCommands.show_subscriptions:
                path_dbg |= 0x00000002
                self._links.log_subscriptions("message")
            case _:
                path_dbg |= 0x00000004
        self.generate_event(
            DBGEvent(Command=dbg, Path=f"0x{path_dbg:08X}", Count=count_dbg, Msg="")
        )
        self._logger.path("--_process_dbg  path:0x%08X  count:%d", path_dbg, count_dbg)

    def add_communicator(self, communicator: CommunicatorInterface) -> None:
        if communicator.name in self._communicators:
            raise ValueError(
                f"ERROR. Communicator with name [{communicator.name}] already present"
            )
        self._communicators[communicator.name] = communicator
        for monitored in communicator.monitored_names:
            self._watchdog.add_monitored_name(monitored)

    async def process_messages(self) -> None:  # noqa: C901
        if self._receive_queue is None:
            raise RuntimeError(
                "ERROR. process_messages() called before Proactor started."
            )
        try:
            self._callbacks.start_processing_messages()
            while not self._stop_requested:
                message = await self._receive_queue.get()
                if not self._stop_requested:
                    await self.async_process_message(message)
                self._receive_queue.task_done()
        except Exception as e:
            if not isinstance(e, asyncio.exceptions.CancelledError):
                self._logger.exception("ERROR in process_message")
                self._logger.error("Stopping proactor")  # noqa: TRY400
                try:
                    self.generate_event(
                        ShutdownEvent(
                            Reason=(
                                f"ERROR in process_message {e}\n"
                                f"{traceback.format_exception(e)}"
                            )
                        )
                    )
                except:  # noqa: E722
                    self._logger.exception("ERROR generating exception event")
        try:
            message_futures = await self._clear_processing_futures()
            for future in message_futures:
                if not future.done():
                    future.cancel()
        except:  # noqa: E722
            ...
        try:
            self.stop()
        except:  # noqa: E722
            self._logger.exception("ERROR stopping proactor")

    def add_task(self, task: asyncio.Task[Any]) -> None:
        self._tasks.append(task)

    def start_tasks(self) -> None:
        self._tasks = [
            asyncio.create_task(self.process_messages(), name="process_messages"),
            *self._links.start_ping_tasks(),
        ]
        self._tasks.extend(self._callbacks.start_tasks())

    @classmethod
    def _second_caller(cls) -> str:
        try:
            # noinspection PyProtectedMember,PyUnresolvedReferences
            return sys._getframe(2).f_back.f_code.co_name  # type: ignore[union-attr] # noqa: SLF001
        except Exception as e:  # noqa: BLE001
            return f"[ERROR extracting caller of _report_errors: {e}"

    def _report_error(self, error: Exception, msg: str = "") -> Result[bool, Exception]:
        try:
            if not msg:
                msg = self._second_caller()
            self._report_errors([error], msg)
        except Exception as e2:  # noqa: BLE001
            return Err(e2)
        return Ok()

    def _report_errors(
        self, errors: Sequence[Exception], msg: str = ""
    ) -> Result[bool, Exception]:
        try:
            if not msg:
                msg = self._second_caller()
            self.generate_event(Problems(errors=errors).problem_event(msg))
        except Exception as e2:  # noqa: BLE001
            return Err(e2)
        return Ok()

    def process_message(self, message: Message[Any]) -> Result[bool, Exception]:
        raise NotImplementedError(
            "Proactor does not implement process_message, "
            "but instead async_process_message."
        )

    async def async_process_message(self, message: Message[Any]) -> None:  # noqa: C901, PLR0912
        if self._logger.path_enabled and not isinstance(message.Payload, PatWatchdog):
            if isinstance(message.Payload, MQTTReceiptPayload):
                msg_type_str = message.Payload.message.topic.split("/")[-1]
            else:
                msg_type_str = f"{message.Header.Src}/{message.Header.MessageType}"
            self._logger.message_enter(
                "++Proactor<%s>.process_message  [%s]", self.short_name, msg_type_str
            )
        path_dbg = 0
        if not isinstance(message.Payload, (MQTTReceiptPayload, PatWatchdog)):
            path_dbg |= 0x00000001
            self._logger.message_summary(
                direction="IN  internal",
                src=message.src(),
                dst=message.dst(),
                topic=f"{message.src()}/to/{message.dst()}/{message.Header.MessageType}",
                payload_object=message.Payload,
                message_id=message.Header.MessageId,
            )
        self._stats.add_message(message)
        match message.Payload:
            case MQTTReceiptPayload():
                path_dbg |= 0x00000002
                self._process_mqtt_message(message)
            case MQTTConnectPayload():
                path_dbg |= 0x00000004
                self._process_mqtt_connected(message)
            case MQTTDisconnectPayload():
                path_dbg |= 0x00000008
                self._process_mqtt_disconnected(message)
            case MQTTConnectFailPayload():
                path_dbg |= 0x00000010
                self._process_mqtt_connect_fail(message)
            case MQTTSubackPayload():
                path_dbg |= 0x00000020
                self._process_mqtt_suback(message)
            case MQTTProblemsPayload():
                path_dbg |= 0x00000040
                self._process_mqtt_problems(message)
            case PatWatchdog():
                path_dbg |= 0x00000080
                self._watchdog.process_message(message)
            case Shutdown():
                path_dbg |= 0x00000100
                self._process_shutdown_message(message)
            case EventBase():
                path_dbg |= 0x00000200
                self.generate_event(message.Payload)
            case _:
                path_dbg |= 0x00000400
                self._callbacks.process_internal_message(message)
        await self._notify_message_future(message)
        if self._logger.path_enabled and not isinstance(message.Payload, PatWatchdog):
            self._logger.message_exit(
                "--Proactor<%s>.process_message  path:0x%08X", self.short_name, path_dbg
            )

    async def _notify_message_future(self, message: Message[Any]) -> None:
        future = getattr(message, self.AWAIT_PROCESSING_FUTURE_ATTRIBUTE, None)
        if future is not None and isinstance(future, asyncio.Future):
            await self._clear_processing_future(future)
            future.set_result(Ok(value=True))

    def _decode_mqtt_message(
        self, mqtt_payload: MQTTReceiptPayload
    ) -> Result[Message[Any], Exception]:
        try:
            result: Result[Message[Any], Exception] = Ok(
                self._links.decode(
                    mqtt_payload.client_name,
                    mqtt_payload.message.topic,
                    mqtt_payload.message.payload,
                )
            )
        except Exception as e:
            self._logger.exception("ERROR decoding [%s]", mqtt_payload)
            clip_len = 70
            self.generate_event(
                ProblemEvent(
                    ProblemType=gwproto.messages.Problems.warning,
                    Summary=f"Decoding error topic [{mqtt_payload.message.topic}]  error [{type(e)}]",
                    Details=(
                        f"Topic: {mqtt_payload.message.topic}\n"
                        f"Message: {mqtt_payload.message.payload[:clip_len]!r}"
                        f"{'...' if len(mqtt_payload.message.payload) > clip_len else ''}\n"
                        f"{traceback.format_exception(e)}\n"
                        f"Exception: {e}"
                    ),
                )
            )
            result = Err(e)
        return result

    def _process_mqtt_message(  # noqa: C901, PLR0912
        self, mqtt_receipt_message: Message[MQTTReceiptPayload]
    ) -> Result[Message[Any], Exception]:
        self._logger.path(
            "++Proactor<%s>._process_mqtt_message %s/%s",
            self.short_name,
            mqtt_receipt_message.Header.Src,
            mqtt_receipt_message.Header.MessageType,
        )
        path_dbg = 0
        self._stats.add_mqtt_message(mqtt_receipt_message)
        match decode_result := self._decode_mqtt_message(mqtt_receipt_message.Payload):
            case Ok(decoded_message):
                path_dbg |= 0x00000001
                decoded_message = decode_result.value
                self._stats.add_decoded_mqtt_message_type(
                    mqtt_receipt_message.Payload.client_name,
                    decoded_message.message_type(),
                )
                if self._logger.message_summary_enabled:
                    if isinstance(decoded_message.Payload, Ack):
                        message_id = decoded_message.Payload.AckMessageID
                    else:
                        message_id = decoded_message.Header.MessageId
                    self._logger.message_summary(
                        direction="IN  mqtt    ",
                        src=decoded_message.src(),
                        dst=decoded_message.dst(),
                        topic=mqtt_receipt_message.Payload.message.topic,
                        payload_object=decoded_message.Payload,
                        message_id=message_id,
                    )
                match self._links.process_mqtt_message(mqtt_receipt_message):
                    case Ok(transition):
                        path_dbg |= 0x00000002
                        if transition.recv_activated():
                            path_dbg |= 0x00000004
                            self._callbacks.recv_activated(transition)
                    case Err(error):
                        path_dbg |= 0x00000008
                        self._report_error(
                            error,
                            "_process_mqtt_message/_link_states.process_mqtt_message",
                        )
                match decoded_message.Payload:
                    case Ack():
                        path_dbg |= 0x00000010
                        self._process_ack(
                            mqtt_receipt_message.Payload.client_name,
                            decoded_message.Payload.AckMessageID,
                        )
                    case Ping():
                        path_dbg |= 0x00000020
                    case DBGPayload():
                        path_dbg |= 0x00000040
                        self._process_dbg(decoded_message.Payload)
                    case _:
                        path_dbg |= 0x00000080
                        self._callbacks.process_mqtt_message(
                            mqtt_receipt_message, decoded_message
                        )
                if decoded_message.Header.AckRequired:
                    path_dbg |= 0x00000200
                    self._links.send_ack(
                        mqtt_receipt_message.Payload.client_name, decoded_message
                    )
        self._logger.path(
            "--Proactor<%s>._process_mqtt_message:%s  path:0x%08X",
            self.short_name,
            int(decode_result.is_ok()),
            path_dbg,
        )
        return decode_result

    def _process_mqtt_connected(self, message: Message[MQTTConnectPayload]) -> None:
        match self._links.process_mqtt_connected(message):
            case Err(error):
                self._report_error(error, "_process_mqtt_connected")

    def _process_mqtt_disconnected(
        self, message: Message[MQTTDisconnectPayload]
    ) -> Result[bool, Exception]:
        result: Result[bool, Exception] = Ok()
        match self._links.process_mqtt_disconnected(message):
            case Ok(transition):
                if transition.recv_deactivated():
                    self._callbacks.recv_deactivated(transition)
                result = Ok(value=True)
            case Err(error):
                result = Err(error)
        return result

    def _process_mqtt_connect_fail(
        self, message: Message[MQTTConnectFailPayload]
    ) -> Result[bool, Exception]:
        result: Result[bool, Exception] = Ok()
        match self._links.process_mqtt_connect_fail(message):
            case Err(error):
                result = Err(error)
        return result

    def _process_mqtt_suback(
        self, message: Message[MQTTSubackPayload]
    ) -> Result[bool, Exception]:
        self._logger.path(
            "++Proactor<%s>._process_mqtt_suback client:%s",
            self.short_name,
            message.Payload.client_name,
        )
        path_dbg = 0
        result: Result[bool, Exception] = Ok()
        match self._links.process_mqtt_suback(message):
            case Ok(transition):
                path_dbg |= 0x00000001
                if transition.recv_activated():
                    path_dbg |= 0x00000002
                    self._callbacks.recv_activated(transition)
                    result = Ok(value=True)
            case Err(error):
                path_dbg |= 0x00000004
                result = Err(error)
        self._logger.path(
            "--Proactor._process_mqtt_suback:%d  path:0x%08X",
            result.is_ok(),
            path_dbg,
        )
        return result

    def _process_mqtt_problems(
        self, message: Message[MQTTProblemsPayload]
    ) -> Result[bool, Exception]:
        self.generate_event(
            ProblemEvent(
                ProblemType=gwproto.messages.Problems.error,
                Summary=f"Error in mqtt event loop for client [{message.Payload.client_name}]",
                Details=(
                    f"{message.Payload.problems}\n"
                    f"{message.Payload.problems.error_traceback_str()}"
                ),
            )
        )
        return Ok()

    def _process_shutdown_message(self, message: Message[Shutdown]) -> None:
        self._stop_requested = True
        self.generate_event(ShutdownEvent(Reason=message.Payload.Reason))
        self._links.flush_in_flight_events()
        self._logger.lifecycle(
            f"Shutting down due to ShutdownMessage, [{message.Payload.Reason}]"
        )

    def _start(self) -> None:
        self._loop = asyncio.get_running_loop()
        self._receive_queue = asyncio.Queue()
        self._links.start(self._loop, self._receive_queue)
        if self._reindex_problems is not None:
            self.generate_event(
                self._reindex_problems.problem_event("Startup event reindex() problems")
            )
        self._reindex_problems = None
        self._callbacks.pre_child_start()
        for communicator in self._communicators.values():
            if isinstance(communicator, Runnable):
                communicator.start()
        self.start_tasks()

    async def run_forever(self) -> None:
        self._start()
        await self.join()

    def run_in_thread(self, *, daemon: bool = True) -> threading.Thread:
        async def _async_run_forever() -> None:
            try:
                await self.run_forever()

            finally:
                self.stop()

        def _run_forever() -> None:
            asyncio.run(_async_run_forever())

        thread = threading.Thread(target=_run_forever, daemon=daemon)
        thread.start()
        return thread

    def start(self) -> None:
        raise RuntimeError("ERROR. Proactor must be started by awaiting run_forever()")

    def stop(self) -> None:
        self._stop_requested = True
        for task in self._tasks:
            if not task.done():
                task.cancel()
        self._links.stop()
        for communicator in self._communicators.values():
            if isinstance(communicator, Runnable):
                try:  # noqa: SIM105
                    communicator.stop()
                except:  # noqa: E722, S110
                    pass

    async def join(self) -> None:
        self._logger.lifecycle("++Proactor<%s>.join()", self.short_name)
        if self._stopped:
            self._logger.lifecycle(
                "--Proactor<%s>.join()  (already stopped)", self.short_name
            )
            return
        if self._loop is None:
            raise ValueError("Proactor cannot be joined until it is started")
        self._logger.lifecycle(str_tasks(self._loop, "Proactor.join() - all tasks"))
        running: set[asyncio.Task[Any]] = set(self._tasks)
        for communicator in self._communicators.values():
            communicator_name = communicator.name
            if isinstance(communicator, Runnable):
                running.add(
                    self._loop.create_task(
                        communicator.join(), name=f"{communicator_name}.join"
                    )
                )

        try:
            while running:
                self._logger.lifecycle(
                    str_tasks(self._loop, "WAITING FOR", tasks=running)
                )
                done, running = await asyncio.wait(
                    running, return_when="FIRST_COMPLETED"
                )
                self._logger.lifecycle(str_tasks(self._loop, tag="DONE", tasks=done))
                self._logger.lifecycle(
                    str_tasks(self._loop, tag="PENDING", tasks=running)
                )
                for task in done:
                    if not task.cancelled() and (exception := task.exception()):
                        self._logger.error(
                            "EXCEPTION in task <%(name)s?>  <%(exception)s>",
                            {
                                "name": task.get_name(),
                                "exception": exception,
                            },
                        )
                        self._logger.error(traceback.format_tb(exception.__traceback__))
            self._stopped = True
        except Exception:
            self._logger.exception("ERROR in Proactor.join")
        self._logger.lifecycle("--Proactor<%s>.join()", self.short_name)



import asyncio
from dataclasses import dataclass
from typing import Any, Callable, Optional, Sequence

from gwproto import Message

from gwproactor.links import Transition
from gwproactor.message import MQTTReceiptPayload

PreChildStartCallback = Callable[[], None]
StartTasksCallback = Callable[[], Sequence[asyncio.Task[Any]]]
StartProcessingMessagesCallback = Callable[[], None]
ProcessMessageCallback = Callable[[Message[Any]], None]
ProcessMQTTMessageCallback = Callable[[Message[MQTTReceiptPayload], Message[Any]], None]
RecvDeactivatedCallback = Callable[[Transition], None]
RecvActivatedCallback = Callable[[Transition], None]


@dataclass
class ProactorCallbackFunctions:
    pre_child_start: Optional[PreChildStartCallback] = None
    start_tasks: Optional[StartTasksCallback] = None
    start_processing_messages: Optional[StartProcessingMessagesCallback] = None
    process_internal_message: Optional[ProcessMessageCallback] = None
    process_mqtt_message: Optional[ProcessMQTTMessageCallback] = None
    recv_activated: Optional[RecvActivatedCallback] = None
    recv_deactivated: Optional[RecvDeactivatedCallback] = None


class ProactorCallbackInterface:
    def pre_child_start(self) -> None: ...
    def start_tasks(self) -> Sequence[asyncio.Task[Any]]:
        return []

    def start_processing_messages(self) -> None: ...
    def process_internal_message(self, message: Message[Any]) -> None: ...
    def process_mqtt_message(
        self, mqtt_client_message: Message[MQTTReceiptPayload], decoded: Message[Any]
    ) -> None: ...
    def recv_activated(self, transition: Transition) -> None: ...
    def recv_deactivated(self, transition: Transition) -> None: ...


class CallbackManager(ProactorCallbackInterface):
    callback_functions: ProactorCallbackFunctions
    callback_objects: dict[int, ProactorCallbackInterface]
    _next_callback_id: int = -1

    def __init__(
        self,
        callback_functions: Optional[ProactorCallbackFunctions] = None,
        callback_objects: Optional[list[ProactorCallbackInterface]] = None,
    ) -> None:
        self.callback_functions = callback_functions or ProactorCallbackFunctions()
        self.callback_objects = {}
        if callback_objects is not None:
            for callback_object in callback_objects:
                self.add_callbacks(callback_object)

    def add_callbacks(self, callbacks: ProactorCallbackInterface) -> int:
        self._next_callback_id += 1
        self.callback_objects[self._next_callback_id] = callbacks
        return self._next_callback_id

    def remove_callbacks(self, callback_id: int) -> None:
        self.callback_objects.pop(callback_id, None)

    def pre_child_start(self) -> None:
        self._call_callbacks("pre_child_start")

    def start_tasks(self) -> Sequence[asyncio.Task[Any]]:
        tasks: list[asyncio.Task[Any]] = []
        if (
            self.callback_functions is not None
            and self.callback_functions.start_tasks is not None
        ):
            tasks.extend(self.callback_functions.start_tasks())
        for callback_object in self.callback_objects.values():
            tasks.extend(callback_object.start_tasks())
        return tasks

    def start_processing_messages(self) -> None:
        self._call_callbacks("start_processing_messages")

    def process_internal_message(self, message: Message[Any]) -> None:
        self._call_callbacks("process_internal_message", message)

    def process_mqtt_message(
        self, mqtt_client_message: Message[MQTTReceiptPayload], decoded: Message[Any]
    ) -> None:
        self._call_callbacks("process_mqtt_message", mqtt_client_message, decoded)

    def recv_activated(self, transition: Transition) -> None:
        self._call_callbacks("recv_activated", transition)

    def recv_deactivated(self, transition: Transition) -> None:
        self._call_callbacks("recv_deactivated", transition)

    def _call_callbacks(self, callback_name: str, *args: Any, **kwargs: Any) -> None:
        if not hasattr(self.callback_functions, callback_name):
            raise RuntimeError(
                f"callback <{callback_name}> is not an attribute of ProactorCallbackFunctions"
            )
        if cb_function := getattr(self.callback_functions, callback_name, None):
            cb_function(*args, **kwargs)
        for callback_object in self.callback_objects.values():
            if (cb_method := getattr(callback_object, callback_name, None)) is None:
                raise RuntimeError(
                    f"callback <{callback_name}> is not an attribute of ProactorCallbackInterface"
                )
            cb_method(*args, **kwargs)



import dataclasses
import time
from dataclasses import dataclass, field
from typing import Optional

from gwproactor.config.proactor_settings import MQTT_LINK_POLL_SECONDS

import_time = time.time()


@dataclass
class LinkMessageTimes:
    last_send: float = field(default_factory=time.time)
    last_recv: float = field(default_factory=time.time)

    def next_ping_second(self, link_poll_seconds: float) -> float:
        return self.last_send + link_poll_seconds

    def seconds_until_next_ping(self, link_poll_seconds: float) -> float:
        return self.next_ping_second(link_poll_seconds) - time.time()

    def time_to_send_ping(self, link_poll_seconds: float) -> bool:
        return time.time() > self.next_ping_second(link_poll_seconds)

    def get_str(
        self,
        *,
        link_poll_seconds: float = MQTT_LINK_POLL_SECONDS,
        relative: bool = True,
    ) -> str:
        adjust = import_time if relative else 0
        return (
            f"n:{time.time() - adjust:5.2f}  lps:{link_poll_seconds:5.2f}  "
            f"ls:{self.last_send - adjust:5.2f}  lr:{self.last_recv - adjust:5.2f}  "
            f"nps:{self.next_ping_second(link_poll_seconds) - adjust:5.2f}  "
            f"snp:{self.next_ping_second(link_poll_seconds):5.2f}  "
            f"tsp:{int(self.time_to_send_ping(link_poll_seconds))}"
        )

    def __str__(self) -> str:
        return self.get_str()


class MessageTimes:
    _links: dict[str, LinkMessageTimes]

    def __init__(self) -> None:
        self._links = {}

    def add_link(self, name: str) -> None:
        self._links[name] = LinkMessageTimes()

    def get_copy(self, link_name: str) -> LinkMessageTimes:
        return dataclasses.replace(self._links[link_name])

    def update_send(self, link_name: str, now: Optional[float] = None) -> None:
        if now is None:
            now = time.time()
        self._links[link_name].last_send = now

    def update_recv(self, link_name: str, now: Optional[float] = None) -> None:
        if now is None:
            now = time.time()
        self._links[link_name].last_recv = now

    def link_names(self) -> list[str]:
        return list(self._links.keys())



# ruff: noqa: S101

import logging
import sys
from types import TracebackType
from typing import Iterator, Optional, Self, Sequence, Type

import pytest

from gwproactor.config import DEFAULT_BASE_NAME, LoggerLevels


class LoggerGuard:
    level: int
    propagate: bool
    disabled: bool
    handlers: set[logging.Handler]
    filters: set[logging.Filter]

    def __init__(self, logger: logging.Logger) -> None:
        self.logger = logger
        self.level = logger.level
        self.propagate = logger.propagate
        self.disabled = logger.disabled
        self.handlers = set(logger.handlers)
        self.filters = set(logger.filters)  # type: ignore[arg-type]

    def restore(self) -> None:
        screen_handlers = [
            h
            for h in self.logger.handlers
            if isinstance(h, logging.StreamHandler)
            and (h.stream is sys.stderr or h.stream is sys.stdout)
        ]
        if len(screen_handlers) > 1:
            raise ValueError(
                "More than 1 screen handlers  "
                f"{self.logger.name}  {len(screen_handlers)}  "
                f"stream handlers: {screen_handlers},  "
                f"from all handlers {self.logger.handlers}"
            )
        self.logger.setLevel(self.level)
        self.logger.propagate = self.propagate
        self.logger.disabled = self.disabled
        curr_handlers = set(self.logger.handlers)
        for handler in curr_handlers - self.handlers:
            self.logger.removeHandler(handler)
        for handler in self.handlers - curr_handlers:
            self.logger.addHandler(handler)
        curr_filters: set[logging.Filter] = set(self.logger.filters)  # type: ignore[arg-type]
        for filter_ in curr_filters - self.filters:
            self.logger.removeFilter(filter_)
        for filter_ in self.filters - curr_filters:
            self.logger.addFilter(filter_)
        assert set(self.logger.handlers) == self.handlers
        assert set(self.logger.filters) == self.filters

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> bool:
        self.restore()
        return True


class LoggerGuards:
    guards: dict[str, LoggerGuard]

    def __init__(self, extra_logger_names: Optional[Sequence[str]] = None) -> None:
        logger_names = self.default_logger_names()
        if extra_logger_names is not None:
            logger_names = logger_names.union(set(extra_logger_names))
        self.guards = {
            logger_name: LoggerGuard(logging.getLogger(logger_name))
            for logger_name in logger_names
        }

    def add_loggers(self, logger_names: Sequence[str]) -> None:
        for logger_name in logger_names:
            if logger_name not in self.guards:
                self.guards[logger_name] = LoggerGuard(logging.getLogger(logger_name))

    def restore(self) -> None:
        for guard in self.guards.values():
            guard.restore()

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> bool:
        self.restore()
        return True

    @classmethod
    def default_logger_names(cls) -> set[str]:
        return {"root"}.union(
            LoggerLevels().qualified_logger_names(DEFAULT_BASE_NAME).values()
        ).union(logging.root.manager.loggerDict.keys())


@pytest.fixture
def restore_loggers() -> Iterator[LoggerGuards]:
    num_root_handlers = len(logging.getLogger().handlers)
    guards = LoggerGuards()
    yield guards
    guards.restore()
    new_num_root_handlers = len(logging.getLogger().handlers)
    if new_num_root_handlers != num_root_handlers:
        raise ValueError(
            f"ARRG. root handlers: {num_root_handlers} -> {new_num_root_handlers}  handlers:\n\t"
            f"{logging.getLogger().handlers}"
        )



import typing
import uuid
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Mapping, Optional

from gwproto import HardwareLayout

from gwproactor.callbacks import ProactorCallbackFunctions
from gwproactor.config.app_settings import AppSettings
from gwproactor.logger import ProactorLogger
from gwproactor.persister import PersisterInterface, StubPersister


@dataclass
class ProactorName:
    long_name: str
    short_name: str

    @cached_property
    def name(self) -> str:
        return self.long_name

    @cached_property
    def publication_name(self) -> str:
        return self.long_name

    @cached_property
    def subscription_name(self) -> str:
        return self.short_name


class ProactorConfig:
    name: ProactorName
    settings: AppSettings
    callback_functions: ProactorCallbackFunctions
    logger: ProactorLogger
    event_persister: PersisterInterface
    layout: HardwareLayout

    def __init__(  # noqa: PLR0913
        self,
        name: ProactorName,
        *,
        settings: Optional[AppSettings] = None,
        callbacks: Optional[ProactorCallbackFunctions] = None,
        logger: Optional[ProactorLogger] = None,
        event_persister: Optional[PersisterInterface] = None,
        hardware_layout: Optional[HardwareLayout] = None,
    ) -> None:
        self.name = name
        self.settings = AppSettings() if settings is None else settings
        self.callback_functions = (
            ProactorCallbackFunctions() if callbacks is None else callbacks
        )
        self.logger = (
            ProactorLogger(
                **typing.cast(
                    Mapping[str, Any], self.settings.logging.qualified_logger_names()
                )
            )
            if logger is None
            else logger
        )
        self.event_persister = (
            StubPersister() if event_persister is None else event_persister
        )
        self.layout = (
            (
                HardwareLayout(
                    layout={
                        "ShNodes": [
                            {
                                "ShNodeId": str(uuid.uuid4()),
                                "Name": name.long_name,
                                "ActorClass": "NoActor",
                                "TypeName": "spaceheat.node.gt",
                            }
                        ]
                    },
                    cacs={},
                    components={},
                    nodes={},
                    data_channels={},
                    synth_channels={},
                )
            )
            if hardware_layout is None
            else hardware_layout
        )



from pathlib import Path
from typing import Any, Optional

import xdg
from pydantic import BaseModel, Field, field_validator
from pydantic_core.core_schema import ValidationInfo

DEFAULT_BASE_NAME: str = "gridworks"
DEFAULT_BASE_DIR = Path(DEFAULT_BASE_NAME)
DEFAULT_NAME: str = "scada"
DEFAULT_NAME_DIR = Path(DEFAULT_NAME)
DEFAULT_LAYOUT_FILE = Path("hardware-layout.json")


class TLSPaths(BaseModel):
    ca_cert_path: Optional[str | Path] = None
    cert_path: Optional[str | Path] = None
    private_key_path: Optional[str | Path] = None

    @classmethod
    def defaults(cls, certs_dir: Path, client_name: str) -> "TLSPaths":
        """Calcuate defaults given a certs_dir and client name. Meant to be called in context where those are known,
        e.g. a validator on a higher-level model which has access to a Paths object and a named MQTT configuration.
        """
        client_dir = Path(certs_dir) / client_name
        return TLSPaths(
            ca_cert_path=client_dir / "ca.crt",
            cert_path=client_dir / f"{client_name}.crt",
            private_key_path=client_dir / "private" / f"{client_name}.pem",
        )

    def effective_paths(self, certs_dir: str | Path, client_name: str) -> "TLSPaths":
        """Re-calculate paths given a certs_dir and client name. Meant to be called in context where those are
        known, e.g. a validator on a higher-level model which has access to a Paths object and a named MQTT
        configuration."""
        fields = self.model_dump(exclude_none=True)
        fields.update(self.defaults(Path(certs_dir), client_name).model_dump())
        return TLSPaths(**fields)

    def mkdirs(
        self, *, mode: int = 0o777, parents: bool = True, exist_ok: bool = True
    ) -> None:
        """Create directories that will be used to store certificates and private keys."""
        if (
            self.ca_cert_path is None
            or self.cert_path is None
            or self.private_key_path is None
        ):
            raise ValueError(
                "ERROR. TLSPaths.mkdirs() requires all paths to nave non-None values. "
                f"Current values: {self.model_dump()}"
            )
        Path(self.ca_cert_path).parent.mkdir(
            mode=mode, parents=parents, exist_ok=exist_ok
        )
        Path(self.cert_path).parent.mkdir(mode=mode, parents=parents, exist_ok=exist_ok)
        Path(self.private_key_path).parent.mkdir(
            mode=mode, parents=parents, exist_ok=exist_ok
        )

    def missing_paths(self) -> list[tuple[str, Optional[Path]]]:
        missing = []
        for path_name in self.__pydantic_fields__:
            path = getattr(self, path_name)
            if path is None or not Path(path).exists():
                missing.append((path_name, path))
        return missing


class Paths(BaseModel):
    # Relative offsets used under home directories
    base: str | Path = Field(default=DEFAULT_BASE_DIR, validate_default=True)
    name: str | Path = Field(default=DEFAULT_NAME_DIR, validate_default=True)
    relative_path: str | Path = Field(default="", validate_default=True)

    # Home directories (defaulting to https://specifications.freedesktop.org/basedir-spec/latest/)
    data_home: str | Path = Field(default="", validate_default=True)
    state_home: str | Path = Field(default="", validate_default=True)
    config_home: str | Path = Field(default="", validate_default=True)

    # Base working paths, defaulting to xdg home/relative_path/...
    data_dir: str | Path = Field(default="", validate_default=True)
    config_dir: str | Path = Field(default="", validate_default=True)
    certs_dir: str | Path = Field(default="", validate_default=True)
    event_dir: str | Path = Field(default="", validate_default=True)
    log_dir: str | Path = Field(default="", validate_default=True)
    hardware_layout: str | Path = Field(default="", validate_default=True)

    @field_validator("base")
    @classmethod
    def get_base(cls, v: str | Path) -> Path:
        return Path(v)

    @field_validator("name")
    @classmethod
    def get_name(cls, v: str | Path) -> Path:
        return Path(v)

    @field_validator("data_home")
    @classmethod
    def get_data_home(cls, v: str | Path) -> Path:
        return Path(v or xdg.xdg_data_home())

    @field_validator("state_home")
    @classmethod
    def get_state_home(cls, v: str | Path) -> Path:
        return Path(v or xdg.xdg_state_home())

    @field_validator("config_home")
    @classmethod
    def get_config_home(cls, v: str | Path) -> Path:
        return Path(v or xdg.xdg_config_home())

    @field_validator("config_dir")
    @classmethod
    def get_config_dir(cls, v: str | Path, info: ValidationInfo) -> Path:
        if not v:
            v = Path(info.data["config_home"]) / info.data["relative_path"]
        return Path(v)

    @field_validator("certs_dir")
    @classmethod
    def get_certs_dir(cls, v: str | Path, info: ValidationInfo) -> Path:
        if not v:
            v = Path(info.data["config_dir"]) / "certs"
        return Path(v)

    @field_validator("data_dir")
    @classmethod
    def get_data_dir(cls, v: str | Path, info: ValidationInfo) -> Path:
        if not v:
            v = Path(info.data["data_home"]) / info.data["relative_path"]
        return Path(v)

    @field_validator("event_dir")
    @classmethod
    def get_event_dir(cls, v: str | Path, info: ValidationInfo) -> Path:
        if not v:
            v = Path(info.data["data_dir"]) / "event"
        return Path(v)

    @field_validator("log_dir")
    @classmethod
    def get_log_dir(cls, v: str | Path, info: ValidationInfo) -> Path:
        if not v:
            v = Path(info.data["state_home"]) / info.data["relative_path"] / "log"
        return Path(v)

    @field_validator("hardware_layout")
    @classmethod
    def get_hardware_layout(cls, v: Any, info: ValidationInfo) -> Path:
        if not v:
            v = Path(info.data["config_dir"]) / DEFAULT_LAYOUT_FILE
        return Path(v)

    @field_validator("relative_path")
    @classmethod
    def get_relative_path(cls, v: str | Path, info: ValidationInfo) -> Path:
        if not v:
            v = Path(info.data["base"]) / info.data["name"]
        return Path(v)

    def mkdirs(
        self, *, mode: int = 0o777, parents: bool = True, exist_ok: bool = True
    ) -> None:
        Path(self.data_dir).mkdir(mode=mode, parents=parents, exist_ok=exist_ok)
        Path(self.config_dir).mkdir(mode=mode, parents=parents, exist_ok=exist_ok)
        Path(self.event_dir).mkdir(mode=mode, parents=parents, exist_ok=exist_ok)
        Path(self.log_dir).mkdir(mode=mode, parents=parents, exist_ok=exist_ok)

    def duplicate(
        self,
        *,
        exclude_unset: bool = True,
        exclude_defaults: bool = True,
        exclude: Optional[set[str]] = None,
        **kwargs: Any,
    ) -> "Paths":
        fields = self.model_dump(
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude=exclude,
        )
        fields.update(**kwargs)
        return Paths(**fields)

    @classmethod
    def default_env_path(
        cls,
        name: str,
        paths: Optional["Paths"] = None,
    ) -> Path:
        paths = Paths(name=name) if paths is None else paths.duplicate(name=name)
        return Path(paths.config_dir).absolute() / ".env"



import asyncio
import typing
from typing import Any, Callable

from gwproactor.links.timer_interface import TimerManagerInterface


class AsyncioTimerManager(TimerManagerInterface):
    def start_timer(
        self, delay_seconds: float, callback: Callable[[], None]
    ) -> asyncio.TimerHandle:
        return asyncio.get_running_loop().call_later(delay_seconds, callback)

    def cancel_timer(self, timer_handle: Any) -> None:
        typing.cast(asyncio.TimerHandle, timer_handle).cancel()



import pytest


def add_live_test_options(parser: pytest.Parser, *, include_tree: bool = False) -> None:
    parser.addoption(
        "--live-test-verbose",
        action="store_true",
        help="Pass verbose=True to LiveTest",
    )
    parser.addoption(
        "--child-verbose",
        action="store_true",
        help="Pass child_verbose=True to LiveTest",
    )
    parser.addoption(
        "--parent-verbose",
        action="store_true",
        help="Pass parent_verbose=True to LiveTest",
    )
    parser.addoption(
        "--parent-on-screen",
        action="store_true",
        help="Pass parent_on_screen=True to LiveTest",
    )
    if include_tree:
        parser.addoption(
            "--child1-verbose",
            action="store_true",
            help="Pass child1_verbose=True to TreeLiveTest",
        )
        parser.addoption(
            "--child2-verbose",
            action="store_true",
            help="Pass child2_verbose=True to TreeLiveTest",
        )
        parser.addoption(
            "--child2-on-screen",
            action="store_true",
            help="Pass child2_on_screen=True to TreeLiveTest",
        )



import abc
from abc import abstractmethod
from typing import Any, Callable


class TimerManagerInterface(abc.ABC):
    """
    Simple interface to infrastructure which can start timers, run callbacks on timer completion, and cancel timers.
    """

    @abstractmethod
    def start_timer(self, delay_seconds: float, callback: Callable[[], None]) -> Any:
        """
        Start a timer. Implementation is expected to call _callback_ after approximately _delay_sceonds_.

        The execution context (e.g. the thread) of the callback must be specified by the implemntation.

        The callback must have sufficient context available to it do its work as well as to detect if it is no longer
        relevant. Note a callback might run after cancelation if the callack was already "in-flight" at time of
        cancellation and it is up to the callback to tolerate this situation.

        Args:
            delay_seconds: The approximate delay before the callback is called.
            callback: The function called after delay_seconds.

        Returns:
            A timer handle which can be passed to _cancel_timer()_ to cancel the callback.
        """

    @abstractmethod
    def cancel_timer(self, timer_handle: Any) -> None:
        """
        Cancel callback associated with _timer_handle_.

        Note that callback might still run after this call returns.

        Args:
            timer_handle: The value returned by start_timer()

        """
